/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "motif.h"
using namespace BALL;
using namespace std;

Motif::Motif()
{
}

/*
 * constructs a Motif object form a definition file and
 * sets all necessary members
 * */
Motif::Motif(String const &motif_definition_file)
{
    // read in the values from the motif definition file
    read_motif_ini_file(motif_definition_file);

    pair_number = 0;

    for (unsigned int i = 0; i < residue_number; i++)
    {
        residue_id_set.insert(residue_ids[i]);
        for (unsigned int j = i + 1; j < residue_number; j++)
        {
            residue_pair_indices[j + i * 10] = pair_number;
            pair<unsigned int, unsigned int> res_pair(i, j);
            residue_pairs.push_back(res_pair);
            ++pair_number;
        }
    }

    // determine the largest interaction center cutoff radius, it will be used for the spacing of the
    // corresponding 3D Hashgrid
    largest_interaction_center_cutoff_radius = 0.0;
    for (unsigned int i = 0; i < interaction_center_cutoff_radii.size(); ++i)
    {
        largest_interaction_center_cutoff_radius
                = (largest_interaction_center_cutoff_radius
                        < interaction_center_cutoff_radii[i])
                        ? interaction_center_cutoff_radii[i]
                        : largest_interaction_center_cutoff_radius;
    }

    // create the map that defines for each amino acid type (not GLY and ALA)
    // the three catalytic atoms in PDB notation
    catalytic_atoms_map = create_catalytic_atoms_map();

    // compute the averages and standard deviations of the 4 geometric variables
    // (CA-CA distance, 2 Angles CB-CA-CA and one diherdal angle CB-CA-CA-CB)
    // for each residue pair in the motif
    compute_motif_distributions();

    // compute the maximal and minimal CA distances for each residue pair
    // that have a probability higher than zero in the CA-normal distribution
    compute_distance_thresholds();
}

Motif::~Motif()
{
}

float const Motif::normal_distribution[350] =
    {0.0000, 0.0040, 0.0080, 0.0120, 0.0160, 0.0190, 0.0239, 0.0279, 0.0319,
     0.0359, 0.0398, 0.0438, 0.0478, 0.0517, 0.0557, 0.0596, 0.0636, 0.0675,
     0.0714, 0.0753, 0.0793, 0.0832, 0.0871, 0.0910, 0.0948, 0.0987, 0.1026,
     0.1064, 0.1103, 0.1141, 0.1179, 0.1217, 0.1255, 0.1293, 0.1331, 0.1368,
     0.1406, 0.1443, 0.1480, 0.1517, 0.1554, 0.1591, 0.1628, 0.1664, 0.1700,
     0.1736, 0.1772, 0.1808, 0.1844, 0.1879, 0.1915, 0.1950, 0.1985, 0.2019,
     0.2054, 0.2088, 0.2123, 0.2157, 0.2190, 0.2224, 0.2257, 0.2291, 0.2324,
     0.2357, 0.2389, 0.2422, 0.2454, 0.2486, 0.2517, 0.2549, 0.2580, 0.2611,
     0.2642, 0.2673, 0.2704, 0.2734, 0.2764, 0.2794, 0.2823, 0.2852, 0.2881,
     0.2910, 0.2939, 0.2969, 0.2995, 0.3023, 0.3051, 0.3078, 0.3106, 0.3133,
     0.3159, 0.3186, 0.3212, 0.3238, 0.3264, 0.3289, 0.3315, 0.3340, 0.3365,
     0.3389, 0.3413, 0.3438, 0.3461, 0.3485, 0.3508, 0.3513, 0.3554, 0.3577,
     0.3529, 0.3621, 0.3643, 0.3665, 0.3686, 0.3708, 0.3729, 0.3749, 0.3770,
     0.3790, 0.3810, 0.3830, 0.3849, 0.3869, 0.3888, 0.3907, 0.3925, 0.3944,
     0.3962, 0.3980, 0.3997, 0.4015, 0.4032, 0.4049, 0.4066, 0.4082, 0.4099,
     0.4115, 0.4131, 0.4147, 0.4162, 0.4177, 0.4192, 0.4207, 0.4222, 0.4236,
     0.4251, 0.4265, 0.4279, 0.4292, 0.4306, 0.4319, 0.4332, 0.4345, 0.4357,
     0.4370, 0.4382, 0.4394, 0.4406, 0.4418, 0.4429, 0.4441, 0.4452, 0.4463,
     0.4474, 0.4484, 0.4495, 0.4505, 0.4515, 0.4525, 0.4535, 0.4545, 0.4554,
     0.4564, 0.4573, 0.4582, 0.4591, 0.4599, 0.4608, 0.4616, 0.4625, 0.4633,
     0.4641, 0.4649, 0.4656, 0.4664, 0.4671, 0.4678, 0.4686, 0.4693, 0.4699,
     0.4706, 0.4713, 0.4719, 0.4726, 0.4732, 0.4738, 0.4744, 0.4750, 0.4756,
     0.4761, 0.4767, 0.4772, 0.4778, 0.4783, 0.4788, 0.4793, 0.4798, 0.4803,
     0.4808, 0.4812, 0.4817, 0.4821, 0.4826, 0.4830, 0.4834, 0.4838, 0.4842,
     0.4846, 0.4850, 0.4854, 0.4857, 0.4861, 0.4864, 0.4868, 0.4871, 0.4875,
     0.4878, 0.4881, 0.4884, 0.4887, 0.4890, 0.4893, 0.4896, 0.4898, 0.4901,
     0.4904, 0.4906, 0.4909, 0.4911, 0.4913, 0.4916, 0.4918, 0.4920, 0.4922,
     0.4925, 0.4927, 0.4929, 0.4931, 0.4932, 0.4934, 0.4936, 0.4938, 0.4940,
     0.4941, 0.4943, 0.4945, 0.4946, 0.4948, 0.4949, 0.4951, 0.4952, 0.4953,
     0.4955, 0.4956, 0.4957, 0.4959, 0.4960, 0.4961, 0.4962, 0.4963, 0.4964,
     0.4965, 0.4966, 0.4967, 0.4968, 0.4969, 0.4970, 0.4971, 0.4972, 0.4973,
     0.4974, 0.4974, 0.4975, 0.4976, 0.4977, 0.4977, 0.4978, 0.4979, 0.4979,
     0.4980, 0.4981, 0.4981, 0.4982, 0.4982, 0.4983, 0.4984, 0.4984, 0.4985,
     0.4985, 0.4986, 0.4986, 0.4987, 0.4987, 0.4987, 0.4988, 0.4988, 0.4989,
     0.4989, 0.4989, 0.4990, 0.4990, 0.4990, 0.4991, 0.4991, 0.4991, 0.4992,
     0.4992, 0.4992, 0.4992, 0.4993, 0.4993, 0.4993, 0.4993, 0.4994, 0.4994,
     0.4994, 0.4994, 0.4994, 0.4995, 0.4995, 0.4995, 0.4995, 0.4995, 0.4995,
     0.4996, 0.4996, 0.4996, 0.4996, 0.4996, 0.4996, 0.4997, 0.4997, 0.4997,
     0.4997, 0.4997, 0.4997, 0.4997, 0.4997, 0.4997, 0.4997, 0.4998};

void Motif::read_motif_ini_file(String const &file_name)
{
    // this set is used to check if the amino acid names in the definition are valid
    StringHashSet amino_acid_types;
    amino_acid_types.insert("HIS");
    amino_acid_types.insert("ASP");
    amino_acid_types.insert("SER");
    amino_acid_types.insert("GLU");
    amino_acid_types.insert("GLN");
    amino_acid_types.insert("ASN");
    amino_acid_types.insert("TRP");
    amino_acid_types.insert("TYR");
    amino_acid_types.insert("LYS");
    amino_acid_types.insert("ARG");
    amino_acid_types.insert("THR");
    amino_acid_types.insert("PHE");
    amino_acid_types.insert("ILE");
    amino_acid_types.insert("LEU");
    amino_acid_types.insert("CYS");
    amino_acid_types.insert("MET");
    amino_acid_types.insert("VAL");

    StringHashSet ids;

    unsigned int source_line_number;
    String value;

    INIFile f(file_name);
    f.read();
    try
    {
        // read the name of the motif
        name = f.getValue("motif", "name");
        if (name == INIFile::UNDEFINED)
            throw String("name");

        value = f.getValue("motif", "residue_number");
        if (value == INIFile::UNDEFINED)
            throw String("residue_number");
        residue_number = value.toUnsignedInt();

        for (unsigned int resi = 1; resi <= residue_number; ++resi)
        {
            String resname = "residue" + String(resi);
            // read in a residue definition
            vector<Vector3> cat_atoms_positions;

            // read in the amino acid type
            value = f.getValue(resname, "type");

            // if the type is not of the 3 letter amino acid code, report an error
            if (!amino_acid_types.has(value))
            {
                source_line_number = __LINE__;
                throw 3;
            }
            // everything seems ok, remember the amino acid type name
            else
                residue_names.push_back(value);

            // read in the residue id
            value = f.getValue(resname, "ID");
            if (value == INIFile::UNDEFINED)
                throw String("ID in section " + resname);

            // if the id is not unique, report an error
            else if (ids.has(value))
            {
                source_line_number = __LINE__;
                throw 4;
            }
            // store the ID
            else
            {
                residue_ids.push_back(value);
                ids.insert(value);
            }

            // read the relative SAS area
            value = f.getValue(resname, "relative_SAS");
            // everything seems ok, so store the SAS area
            if (value ==  INIFile::UNDEFINED)
                relative_SAS.push_back(-1.0);
            else
                relative_SAS.push_back(value.toDouble());

            for (unsigned int atomi = 1; atomi <= 3; ++atomi)
            {
                value = f.getValue(resname, "atom" + String(atomi));
                // if there are not exactly three floating point coordinates
                // for the atom positions, report an error
                if (!value.countFields() == 3
                        || !value.getField(0).isFloat()
                        || !value.getField(1).isFloat()
                        || !value.getField(2).isFloat())
                {
                    source_line_number = __LINE__;
                    throw 6;
                }
                // store the atom position
                else
                {
                    Vector3 position(value.getField(0).toDouble(),
                                     value.getField(1).toDouble(),
                                     value.getField(2).toDouble());
                    cat_atoms_positions.push_back(position);
                }
            }
            catalytic_atom_positions.push_back(cat_atoms_positions);
            catalytic_atom_positions_linear.insert(
                                        catalytic_atom_positions_linear.end(),
                                        cat_atoms_positions.begin(),
                                        cat_atoms_positions.end());

        }

        value = f.getValue("motif", "interaction_center_number");
        if (value != INIFile::UNDEFINED)
        {
            unsigned int ia_number = value.toUnsignedInt();
            for (unsigned int iai = 1; iai <= ia_number; ++iai)
            {
                String ianame = "interaction_center" + String(iai);
                String pos = f.getValue(ianame, "position");

                // if there are not exactly three floating point coordinates
                // for the ia positions, report an error
                if (!pos.countFields() == 3
                        || !pos.getField(0).isFloat()
                        || !pos.getField(1).isFloat()
                        || !pos.getField(2).isFloat())
                {
                    source_line_number = __LINE__;
                    throw 6;
                }
                // store the ia position
                else
                {
                    Vector3 position(pos.getField(0).toDouble(),
                                     pos.getField(1).toDouble(),
                                     pos.getField(2).toDouble());
                    interaction_center_positions.push_back(position);
                }
                String cutoffRadius = f.getValue(ianame, "cutoff_radius");
                if (cutoffRadius == INIFile::UNDEFINED)
                    throw String("cutoff_radius in section " + ianame);
                interaction_center_cutoff_radii.push_back
                                                    (cutoffRadius.toDouble());

                String expression = f.getValue(ianame, "expression");
                if (expression == INIFile::UNDEFINED)
                    throw String("expression in section " + ianame);
                Expression ex(expression);
                interaction_center_expressions.push_back(ex);
            }
        }

        value = f.getValue("thresholds", "probability_threshold");
        if (value == INIFile::UNDEFINED)
            throw String("probability_threshold");
        probability_threshold = value.toDouble();

        value = f.getValue("thresholds","maximal_backbone_clash_penalty");
        if (value == INIFile::UNDEFINED)
            throw String("maximal_backbone_clash_penalty");
        maximal_backbone_clash_penalty = value.toFloat();

        value = f.getValue("thresholds","maximal_catalytic_geometry_penalty");
        if (value == INIFile::UNDEFINED)
            throw String("maximal_catalytic_geometry_penalty");
        maximal_catalytic_geometry_penalty = value.toFloat();

        // read the rotamer library file
        rotamer_lib_file = f.getValue("paths", "rotamer_library_file");
        if (value == INIFile::UNDEFINED)
            throw String("rotamer_library_file");

        value = f.getValue("paths", "substrate_pdb_file");
        if (value != INIFile::UNDEFINED)
        {
            substrate_system = System();
            PDBFile pdb_file;
            pdb_file.open(value);
            if (!pdb_file.hasFormat())
            {
                source_line_number = __LINE__;
                throw 9;
            }
            pdb_file.read(substrate_system);
            if (!substrate_system.isValid())
            {
                source_line_number = __LINE__;
                throw 9;
            }
        }
        // if there are to few residues, report it
        if (residue_number == 1)
        {
            source_line_number = __LINE__;
            throw 2;
        }
    }
    catch (int a)
    {
        String message = "Error in motif definition file in line:\n";
        if (a == 2)
        {
            message += "Motif must contain at least two residues.";
        }
        else if (a == 3)
        {
            message
                    += "Residue type must be an amino acid 3 letter code. "
                            "(GLY, ALA and PRO are not supported).";
        }
        else if (a == 4)
        {
            message += "Residue IDs have to be unique.";
        }
        else if (a == 6)
        {
            message
                    += "An atom position must contain exactly three "
                            "floating point coordinates.";
        }
        else if (a == 9)
        {
            message += "The substrate PDB file has invalid format.";
        }
        Exception::GeneralException e(__FILE__, source_line_number);
        e.setMessage(message);
        throw e;
    }
    catch (String s)
    {
        String message = "Motif file parameter \"" + s + "\" is missing.";
        Exception::GeneralException e(file_name.c_str(), -1,
                                      "Motif file", message);
        throw e;
    }
}



/*
 * returns the amino acid type of residue i
 */
const String& Motif::get_residue_name(unsigned int i) const
{
    return residue_names[i];
}

/*
 * returns the ID of Residue i
 */
const String& Motif::get_residue_id(unsigned int i) const
{
    return residue_ids[i];
}

/*
 * returns a set with all ids of residues in the motif
 */
const HashSet<String>& Motif::get_residue_id_set() const
{
    return residue_id_set;
}

/*
 * returns minimal CA distance that has a probability > threshold for at least one residue pair
 */
const double Motif::get_minimal_CA_distance() const
{
    return minimal_CA_distance;
}

/*
 * returns maximal CA distance that has a probability > threshold for at least one residue pair
 */
const double Motif::get_maximal_CA_distance() const
{
    return maximal_CA_distance;
}

/*
 * returns the probability threshold
 */
const double Motif::get_probability_threshold() const
{
    return probability_threshold;
}

const float Motif::get_maximal_backbone_clash_penalty() const
{
    return maximal_backbone_clash_penalty;
}

const float Motif::get_maximal_catalytic_geometry_penalty() const
{
    return maximal_catalytic_geometry_penalty;
}


/*
 * returns number of residue pairs in motif
 */
const unsigned int Motif::get_residue_pair_count() const
{
    return pair_number;
}

/*
 * returns number of residues in motif
 */
const unsigned int Motif::get_residue_count() const
{
    return residue_number;
}

/*
 * returns relative SAS area for residue i
 */
const double Motif::get_relative_SAS_for_residue(unsigned int i) const
{
    return relative_SAS[i];
}

/*
 * returns a pointer to built rotamer rot of residue res
 * (built rotamer = a residue with the chi torsions of the rotamer rot)
 * (cannot be const, because the residue is rotated and translated in motifsearch)
 */
Residue* Motif::get_rotamer(unsigned int res, unsigned int rot)
{
    return &rotamers[res][rot];
}

/*
 * returns returns the number of rotamers in the library for residue i
 */
const unsigned int Motif::get_number_of_rotamers_for_residue(unsigned int i) const
{
    return rotamers[i].size();
}

/*
 * returns the PDB name of catalytic atom number "atom" in residue no. res
 */
const String& Motif::get_catalytic_atom_name(unsigned int res,
                                             unsigned int atom) const
{
    return catalytic_atom_names[res][atom];
}

/*
 * returns the position of catalytic atom no. "atom" in residue res
 */
const Vector3& Motif::get_catalytic_atom_position(unsigned int res,
                                                  unsigned int atom) const
{
    return catalytic_atom_positions[res][atom];
}

/*
 * returns the positions of all catalytic atoms in order as defined in the motif file
 */
const vector<Vector3>& Motif::get_catalytic_atom_positions_vector() const
{
    return catalytic_atom_positions_linear;
}

/*
 * returns the frequency of the rotamer no. rot for residue no. res according to the rotamer library
 */
const double Motif::get_rotamer_frequency(unsigned int res, unsigned int rot) const
{
    return rotamer_frequencies[res][rot];
}

/*
 * returns the indices in the residues vector of motif residue pair i
 */
const pair<unsigned int, unsigned int> Motif::get_residue_pair(unsigned int i) const
{
    return residue_pairs[i];
}

/**
 * returns the index if the motif residue pair formed by motif residues i & j
 */
const unsigned int Motif::get_residue_pair_index(unsigned int i, unsigned int j) const
{
    return residue_pair_indices[10* i + j];
}

/*
 * returns the averages of geometric values CA-distance, CB1-CA1-CA2-angle, CA1-CA2-CB2-angle,
 * CB1-CA1-CA2-CB2 dihedral angle of residue pair i in that order
 */
const vector<double>& Motif::get_averages_for_residue_pair(unsigned int i) const
{
    return averages[i];
}

/*
 * returns the standard deviations of geometric values CA-distance, CB1-CA1-CA2-angle, CA1-CA2-CB2-angle,
 * CB1-CA1-CA2-CB2 dihedral angle of residue pair i in that order
 */
const vector<double>& Motif::get_deviations_for_residue_pair(unsigned int i) const
{
    return deviations[i];
}

/*
 * returns the distance between catalytic atom a1 in residue res1 and catalytic atom a2 in res2
 */
const double Motif::get_catalytic_atom_distance(unsigned int res1,
                                                unsigned int res2,
                                                unsigned int a1,
                                                unsigned int a2) const
{
    return catalytic_atom_distances[res1][res2][a1][a2];
}

/*
 * returns the torsion angle of atoms a11 , a12 of residue res1 and atoms a21, a22 of residue res2
 * where a11 < a12, a21 < a22, res1 < res2
 */
const double Motif::get_catalytic_atom_torsion(unsigned int res1,
                                               unsigned int res2,
                                               unsigned int a11,
                                               unsigned int a12,
                                               unsigned int a21,
                                               unsigned int a22) const
{
    unsigned int key = res1 + 100 * res2 + 10000 * a11 + 100000 * a12 + 1000000
            * a21 + 10000000 * a22;
    if (catalytic_atom_torsions.has(key))
    {
        return catalytic_atom_torsions[key];
    }
    else
        return -10;
}

/*
 * returns the BALL expression of interaction center i
 */
const Expression Motif::get_interaction_center_expression(unsigned int i) const
{
    return interaction_center_expressions[i];
}

/*
 * returns the position of interaction center i
 */
const Vector3 Motif::get_interaction_center_position(unsigned int i) const
{
    return interaction_center_positions[i];
}

/*
 * returns the cutoff radius of interaction center i
 */
const double Motif::get_interaction_center_cutoff_radius(unsigned int i) const
{
    return interaction_center_cutoff_radii[i];
}

/*
 * returns the number of interaction centers
 */
const unsigned int Motif::get_interaction_center_count() const
{
    return interaction_center_expressions.size();
}

/*
 * returns the distance between catalytic atom a1 in residue res1 and catalytic atom a2 in res2
 */
const double Motif::get_maximal_interaction_cutoff_radius() const
{
    return largest_interaction_center_cutoff_radius;
}

/*
 * returns the name of the motif
 */
const String Motif::get_name() const
{
    return name;
}

/*
 * returns a pointer to the ligand pdb system;
 */
System* Motif::get_substrate_system()
{
    return &substrate_system;
}

/*
 * computes the distributions of the geometric variables (CA-distance, 3 Angles)
 * for all residue pairs (i.e. averages and standard deviations)
 * Addidtionally, the distances of all catalytic atom pairs that are
 * not in the same residue are computed and stored.
 * Torsion angles are computed for each combination of 2 catalytic atom positions
 * per for 2 residues
 * (Each torsion is computed for 4 positions. two of them are derived from one residue, two from the
 * other. this is doen for each residue pair, and for all possible atom pair combinations)
 */
void Motif::compute_motif_distributions()
{
    // Open fragment database for rotamer library
    FragmentDB fragment_db("");

    // use rotamer library in dunbrack format
    RotamerLibrary rot_lib(rotamer_lib_file, fragment_db);

    // loop over all residues of the motif
    for (unsigned int i = 0; i < residue_number; ++i)
    {
        // get the rotamers for the current residue
        ResidueRotamerSet* rotamer_set =
                rot_lib.getRotamerSet(residue_names[i]);

        vector<Residue> built_rotamers;
        vector<double> frequencies;

        catalytic_atom_names.push_back(catalytic_atoms_map[residue_names[i]]);

        // if there are no rotamers, remember the motif residue as only "rotamer" with a frequency of 100 %
        if (!rotamer_set)
        {
            Exception::GeneralException e;
            String message =
                    "Could not find a rotamer set in library for residue "
                            + residue_names[i];
            e.setMessage(message);
            throw e;
        }
        else
        {

            // build all rotamers for the current residue (create a Residue with the rotamers chi angles),
            // store the built rotamers and their frequencies
            for (unsigned int j = 0; j < rotamer_set->getNumberOfRotamers(); ++j)
            {
                // This is a ugly hack:
                // ResidueRotamerSet cannot assign rotamers/angles to a residue at a terminal
                // --> create a peptide with the residue of interest enclosed between two glycins
                String seq = "G"
                        + String(Peptides::OneLetterCode(residue_names[i]))
                        + "G";
                Peptides::PeptideBuilder pb(seq);
                FragmentDB fragment_db("");
                pb.setFragmentDB(&fragment_db);
                Protein* prot = pb.construct();
                Residue r;

                // select the second residue in our dummy peptide, i.e. the one we are interested in
                // and assign the current rotamer's sidechain angles
                r = *(++(prot->beginResidue()));
                Rotamer rotamer = rotamer_set->getRotamer(j);


                rotamer_set->setRotamer(r, rotamer);
                built_rotamers.push_back(r);

                // halve the frequency, because the number of ASP/GLU rotamers will be doubled
                if (residue_names[i] == "ASP" || residue_names[i] == "GLU")
                {
                    frequencies.push_back(0.5 * rotamer.P);
                }
                else
                    frequencies.push_back(rotamer.P);
            }
            // For each ASP/GLU rotamer of the original library, create another
            // rotamer with the positions of OD1/OE1 and OD2/OE2 swapped, because they are chemically equivalent
            if (residue_names[i] == "ASP" || residue_names[i] == "GLU")
            {
                for (unsigned int j = 0; j < rotamer_set->getNumberOfRotamers(); ++j)
                {
                    String seq = "G"
                            + String(Peptides::OneLetterCode(residue_names[i]))
                            + "G";
                    Peptides::PeptideBuilder pb(seq);
                    FragmentDB fragment_db("");
                    pb.setFragmentDB(&fragment_db);
                    Protein* prot = pb.construct();
                    Residue r;
                    r = *(++(prot->beginResidue()));
                    Rotamer rotamer = rotamer_set->getRotamer(j);
                    rotamer_set->setRotamer(r, rotamer);
                    if (residue_names[i] == "ASP")
                    {
                        Vector3 swap = r.getAtom("OD1")->getPosition();
                        r.getAtom("OD1")->setPosition(
                                             r.getAtom("OD2")->getPosition());
                        r.getAtom("OD2")->setPosition(swap);
                    }
                    else
                    {
                        Vector3 swap = r.getAtom("OE1")->getPosition();
                        r.getAtom("OE1")->setPosition(
                                             r.getAtom("OE2")->getPosition());
                        r.getAtom("OE2")->setPosition(swap);
                    }
                    built_rotamers.push_back(r);
                    frequencies.push_back(0.5 * rotamer.P);
                }
            }
        }
        rotamers.push_back(built_rotamers);
        rotamer_frequencies.push_back(frequencies);
    }
    vector<vector<Residue> > backup_rotamers(rotamers);

    unsigned int i1, i2, j1, j2;

    // in catalytic_atom_distances[i][j][k][l] is the distance of catalytic atom #k in residue i and catalytic atom #l in residue j
    // loop over the motif residues
    for (i1 = 0; i1 < catalytic_atom_positions.size(); ++i1)
    {
        vector<vector<vector<double> > > v1;

        // loop over the motif residues
        for (i2 = 0; i2 < catalytic_atom_positions.size(); ++i2)
        {
            vector<vector<double> > v2;

            // loop over the 3 catalytic atoms
            for (j1 = 0; j1 < catalytic_atom_positions[i1].size(); ++j1)
            {
                vector<double> v3;

                // loop over the 3 catalytic atoms
                for (j2 = 0; j2 < catalytic_atom_positions[i2].size(); ++j2)
                {
                    // store the distance between the two catalytic atoms
                    v3.push_back((catalytic_atom_positions[i1][j1]
                            - catalytic_atom_positions[i2][j2]).getLength());
                }
                v2.push_back(v3);
            }
            v1.push_back(v2);
        }
        catalytic_atom_distances.push_back(v1);
    }

    // loop over all residue pairs in the motif for computing the distribution of the geometric values
    for (i1 = 0; i1 < residue_number; ++i1)
    {
        for (i2 = i1 + 1; i2 < residue_number; ++i2)
        {
            vector<double> pair_averages(4, 0.0), pair_deviations;

            // compute the geometric values using function "compute pairwise distributions"
            compute_pairwise_distribution(i1, i2, pair_averages,
                                          pair_deviations);
            averages.push_back(pair_averages);
            deviations.push_back(pair_deviations);
        }
    }

    vector<pair<unsigned int, unsigned int> >::iterator pit;

    // loop over all motif residue pairs
    for (pit = residue_pairs.begin(); pit != residue_pairs.end(); ++pit)
    {
        // oop over catalytic atom pairs for residue 1 of the pair
        for (i1 = 0; i1 < 2; ++i1)
        {
            for (i2 = i1 + 1; i2 < 3; ++i2)
            {
                // loop over catalytic atom pairs for residue 2 of the pair
                for (j1 = 0; j1 < 2; ++j1)
                {
                    for (j2 = j1 + 1; j2 < 3; ++j2)
                    {
                        // compute the torsion angle of the 4 atoms
                        double
                                torsion =
                                        getTorsionAngle(
                                                        catalytic_atom_positions[pit->first][i1].x,
                                                        catalytic_atom_positions[pit->first][i1].y,
                                                        catalytic_atom_positions[pit->first][i1].z,
                                                        catalytic_atom_positions[pit->first][i2].x,
                                                        catalytic_atom_positions[pit->first][i2].y,
                                                        catalytic_atom_positions[pit->first][i2].z,
                                                        catalytic_atom_positions[pit->second][j1].x,
                                                        catalytic_atom_positions[pit->second][j1].y,
                                                        catalytic_atom_positions[pit->second][j1].z,
                                                        catalytic_atom_positions[pit->second][j2].x,
                                                        catalytic_atom_positions[pit->second][j2].y,
                                                        catalytic_atom_positions[pit->second][j2].z);

                        // this hashkey codes the pair of residues and the catalytic atoms
                        // that were used to compute the torsion angle
                        // example: key: 2 1 2 0 01 00 is derived from residue 0, atoms 0 and 2, and residue 1, atoms 1 and 2
                        unsigned int key = pit->first + 100 * pit->second
                                + 10000 * i1 + 100000* i2 + 1000000* j1
                                + 10000000* j2 ;
                        catalytic_atom_torsions[key] = torsion;
                    }
                }
            }
        }
    }
    rotamers = backup_rotamers;
}

/*
 * Compute the distribution of the Ca-Cb distances & angles of residues res1 & res2.
 * The catalytic atoms are kept fix, all built rotamers are mapped onto the catalytic atoms
 * (i.e. compute a transformation matrix that maps the cat. atom positions in a rotamer onto the positions
 * given in the motif definitions, and apply that matrix to all atom positions in the built rotamer)
 * Use the rotamers to compute the distributions of the Ca-Ca distance & Ca/Cb angles.
 */
void Motif::compute_pairwise_distribution(unsigned int index1,
                                          unsigned int index2,
                                          vector<double>& averages, vector<
                                                  double>& deviations)
{
    Matrix4x4 T;
    StructureMapper mapper;

    Vector3 v1, v2, v3;

    // all computed vectors [Ca-distance, phi, psi, tau] will be stored here
    // so we can compute the average and standard deviation of each value
    vector<vector<double> > all_geometric_values;
    unsigned int i1, i2;

    vector<Vector3> CA1s, CA2s, CB1s, CB2s;

    if (residue_names[index1] == "SER" || residue_names[index1] == "CYS")
    {
        bool is_ser = (residue_names[index1] == "SER");
        CA1s = vector<Vector3> (3);
        CB1s = vector<Vector3> (3);
        CA1s[0] = catalytic_atom_positions[index1][0];
        CB1s[0] = catalytic_atom_positions[index1][1];
        pair<Vector3, Vector3> p =
            scsel_util::get_rotamer_ca_positions_for_ser_cys(CA1s[0], CB1s[0],
                                           catalytic_atom_positions[index1][2],
                                           is_ser);
        CA1s[1] = p.first;
        CB1s[1] = CB1s[0];
        CA1s[2] = p.second;
        CB1s[2] = CB1s[0];
    }
    else
    {
        CA1s = vector<Vector3> (rotamers[index1].size());
        CB1s = vector<Vector3> (rotamers[index1].size());
        // loop over all rotamers of res1
        for (i1 = 0; i1 < rotamers[index1].size(); ++i1)
        {
            Residue r = rotamers[index1][i1];
            // get the positions of the catalytic atom in the current rotamer of res1
            v1
                    = rotamers[index1][i1].getAtom(
                                                   catalytic_atom_names[index1][0])->getPosition();
            v2
                    = rotamers[index1][i1].getAtom(
                                                   catalytic_atom_names[index1][1])->getPosition();
            v3
                    = rotamers[index1][i1].getAtom(
                                                   catalytic_atom_names[index1][2])->getPosition();

            // superimpose the three catalytic atoms of the rotamer onto the catalytic atoms of res1
            // by this moving the rotamer into the right position w.r.t the enzymatic motif
            T = mapper.matchPoints(v1, v2, v3,
                                   catalytic_atom_positions[index1][0],
                                   catalytic_atom_positions[index1][1],
                                   catalytic_atom_positions[index1][2]);

            mapper.setTransformation(T);
            rotamers[index1][i1].apply(mapper);

            CA1s[i1] = rotamers[index1][i1].getAtom("CA")->getPosition();
            if (rotamers[index1][i1].getName() == "GLY")
            {
                CB1s[i1]
                        = scsel_util::get_ideal_cb_position_for_gly(
                                                        rotamers[index1][i1].getAtom(
                                                                                     "N")->getPosition(),
                                                        rotamers[index1][i1].getAtom(
                                                                                     "C")->getPosition(),
                                                        CA1s[i1]);
            }
            else
                CB1s[i1] = rotamers[index1][i1].getAtom("CB")->getPosition();
        }
    }

    if (residue_names[index2] == "SER" || residue_names[index2] == "CYS")
    {
        bool is_ser = (residue_names[index2] == "SER");
        CA2s = vector<Vector3> (3);
        CB2s = vector<Vector3> (3);
        CA2s[0] = catalytic_atom_positions[index2][0];
        CB2s[0] = catalytic_atom_positions[index2][1];
        pair<Vector3, Vector3>
                p =
                    scsel_util::get_rotamer_ca_positions_for_ser_cys(
                                                             CA2s[0],
                                                             CB2s[0],
                                                             catalytic_atom_positions[index2][2],
                                                             is_ser);
        CA2s[1] = p.first;
        CB2s[1] = CB2s[0];
        CA2s[2] = p.second;
        CB2s[2] = CB2s[0];
    }
    else
    {
        CA2s = vector<Vector3> (rotamers[index2].size());
        CB2s = vector<Vector3> (rotamers[index2].size());
        // loop over all rotamers of res2
        for (i2 = 0; i2 < rotamers[index2].size(); ++i2)
        {
            // get the positions of the catalytic atom in the current rotamer of res2
            v1
                    = rotamers[index2][i2].getAtom(
                                                   catalytic_atom_names[index2][0])->getPosition();
            v2
                    = rotamers[index2][i2].getAtom(
                                                   catalytic_atom_names[index2][1])->getPosition();
            v3
                    = rotamers[index2][i2].getAtom(
                                                   catalytic_atom_names[index2][2])->getPosition();

            // superimpose the three catalytic atoms of the rotamers onto the catalytic atoms of res2
            // by this moving the rotamer into the right position w.r.t the enzymatic motif
            T = mapper.matchPoints(v1, v2, v3,
                                   catalytic_atom_positions[index2][0],
                                   catalytic_atom_positions[index2][1],
                                   catalytic_atom_positions[index2][2]);
            mapper.setTransformation(T);
            rotamers[index2][i2].apply(mapper);

            // compute the geometric values (Ca-distance, 3 angles) of the 2 current rotamers using function
            // "compute_geometric_values"
            CA2s[i2] = rotamers[index2][i2].getAtom("CA")->getPosition();
            if (rotamers[index2][i2].getName() == "GLY")
            {
                CB2s[i2]
                        = scsel_util::get_ideal_cb_position_for_gly(
                                                        rotamers[index2][i2].getAtom(
                                                                                     "N")->getPosition(),
                                                        rotamers[index2][i2].getAtom(
                                                                                     "C")->getPosition(),
                                                        CA2s[i2]);
            }
            else
                CB2s[i2] = rotamers[index2][i2].getAtom("CB")->getPosition();
        }
    }

    //compute the distributions by looping over the CA-CB vector pairs
    for (i1 = 0; i1 < CA1s.size(); ++i1)
    {
        for (i2 = 0; i2 < CA2s.size(); ++i2)
        {
            vector<double> geometric_values =
                scsel_util::compute_geometric_values(CA1s[i1], CB1s[i1], CA2s[i2],
                                             CB2s[i2]);
            double freq1, freq2;
            if (residue_names[index1] == "SER" || residue_names[index1]
                    == "CYS")
            {
                freq1 = 0.3333;
            }
            else
                freq1 = rotamer_frequencies[index1][i1];
            if (residue_names[index2] == "SER" || residue_names[index2]
                    == "CYS")
            {
                freq2 = 0.3333;
            }
            else
                freq2 = rotamer_frequencies[index2][i2];

            // compute the joint frequency of the rotamer pair
            double frequency = freq1 * freq2;
            for (unsigned int i = 0; i < 4; ++i)
            {
                averages[i] += geometric_values[i] * frequency;
            }
            // remember the frequency of the rotamer pair in the 5th field of geometric_values
            geometric_values.push_back(frequency);

            // remember the values, so we can compute the standard deviation later
            all_geometric_values.push_back(geometric_values);
        }
    }

    vector<vector<double> >::iterator vit;

    // compute the standard deviations of the Ca-distance and the 3 angles
    double dev = 0.0;
    if (all_geometric_values.size() > 1)
    {
        double correction = (double)all_geometric_values.size()
                / ((double)all_geometric_values.size() - 1.0);
        for (int i = 0; i < 4; ++i)
        {
            for (vit = all_geometric_values.begin(); vit
                    != all_geometric_values.end(); ++vit)
            {
                // weigh the squared differences of the geometric values by the frequency (in vit[4])
                // of the rotamer pair that they were measured in
                dev += pow((*vit)[i] - averages[i], 2) * (*vit)[4];
            }
            // correct the standard deviation with the factor of n/n-1 (to match the standard deviation formula)
            dev *= correction;
            dev = sqrt(dev);
            deviations.push_back(dev);
        }
    }
}

/**
 * creates a map that assigns each amino acid type (save GLY and ALA)
 * the names of its three catalytic atoms in PDB notation in a defined order
 */
map<String, vector<String> > Motif::create_catalytic_atoms_map()
{
    map<String, vector<String> > aa_to_atoms;
    vector<String> v1;
    v1.push_back(String("CB"));
    v1.push_back(String("CG1"));
    v1.push_back(String("CG2"));
    aa_to_atoms["VAL"] = v1;

    vector<String> v2;
    v2.push_back(String("CG"));
    v2.push_back(String("CD1"));
    v2.push_back(String("CD2"));
    aa_to_atoms["PHE"] = v2;

    vector<String> v4;
    v4.push_back(String("CG"));
    v4.push_back(String("SD"));
    v4.push_back(String("CE"));
    aa_to_atoms["MET"] = v4;

    vector<String> v5;
    v5.push_back(String("CB"));
    v5.push_back(String("CG1"));
    v5.push_back(String("CD1"));
    aa_to_atoms["ILE"] = v5;

    vector<String> v6;
    v6.push_back(String("CG"));
    v6.push_back(String("CD1"));
    v6.push_back(String("CD2"));
    aa_to_atoms["LEU"] = v6;

    vector<String> v7;
    v7.push_back(String("CG"));
    v7.push_back(String("OD1"));
    v7.push_back(String("OD2"));
    aa_to_atoms["ASP"] = v7;

    vector<String> v8;
    v8.push_back(String("CD"));
    v8.push_back(String("OE1"));
    v8.push_back(String("OE2"));
    aa_to_atoms["GLU"] = v8;

    vector<String> v9;
    v9.push_back(String("CD"));
    v9.push_back(String("CE"));
    v9.push_back(String("NZ"));
    aa_to_atoms["LYS"] = v9;

    vector<String> v10;
    v10.push_back(String("CZ"));
    v10.push_back(String("NH1"));
    v10.push_back(String("NH2"));
    aa_to_atoms["ARG"] = v10;

    vector<String> v11;
    v11.push_back(String("CA"));
    v11.push_back(String("CB"));
    v11.push_back(String("OG"));
    aa_to_atoms["SER"] = v11;

    vector<String> v12;
    v12.push_back(String("CB"));
    v12.push_back(String("OG1"));
    v12.push_back(String("CG2"));
    aa_to_atoms["THR"] = v12;

    vector<String> v13;
    v13.push_back(String("OH"));
    v13.push_back(String("CE1"));
    v13.push_back(String("CE2"));
    aa_to_atoms["TYR"] = v13;

    vector<String> v14;
    v14.push_back(String("NE2"));
    v14.push_back(String("ND1"));
    v14.push_back(String("CE1"));
    aa_to_atoms["HIS"] = v14;

    vector<String> v15;
    v15.push_back(String("CA"));
    v15.push_back(String("CB"));
    v15.push_back(String("SG"));
    aa_to_atoms["CYS"] = v15;

    vector<String> v16;
    v16.push_back(String("CG"));
    v16.push_back(String("OD1"));
    v16.push_back(String("ND2"));
    aa_to_atoms["ASN"] = v16;

    vector<String> v17;
    v17.push_back(String("CD"));
    v17.push_back(String("OE1"));
    v17.push_back(String("NE2"));
    aa_to_atoms["GLN"] = v17;

    vector<String> v18;
    v18.push_back(String("NE1"));
    v18.push_back(String("CD1"));
    v18.push_back(String("CD2"));
    aa_to_atoms["TRP"] = v18;

    return aa_to_atoms;
}

double Motif::get_normal_probability(double d) const
{
    d = fabs(d);
    int i = (int)floor(100.0 * d);
//    return normal_distribution_vector[i];
    return normal_distribution[i];
}


/*
 * Looks in the distributions of the Ca distances of the motif for the
 * maximal and the minimal Ca-distance that can theoretically score
 * a probability higher than zero. these two values will be used to
 * determine which residue pairs are to be considered for pair hits
 */
void Motif::compute_distance_thresholds()
{
    double deviation_mult = 1.0;

    // if the probability threshold is smaller then the smallest probability in the probability table
    // (that is the probability for a value of 3.49 in the standard normal distribution)
    // use 3.5 as the maximal standard deviation
    if (probability_threshold < (1.0 - 2* normal_distribution[349]))
    {
        deviation_mult = 3.5;
    }
    // check for the biggest value for which the probability is higher than the threshold,
    // remember it as the highest permitted deviation
    else
    {
        for (unsigned int i = 349; i >= 0; --i)
        {
            if (probability_threshold < (1.0 - 2*
                    normal_distribution[i]))
            {
                deviation_mult = i / 100.0 + 0.01;
                break;
            }
        }
    }

    minimal_CA_distance = -1.0;
    maximal_CA_distance = -1.0;
    vector<vector<double> >::iterator it = averages.begin();

    // this loop checks for the highest average Ca distance of all motif residue pairs, and adds to it the
    // standard deviation of this Ca-distance distribution times the deviation multiplication factor
    // calculated above
    // similarly, it also checks for the lowest Ca distance, and subtracts the standard deviation
    // multiplicated with the deviation multiplication factor
    for (unsigned int i = 0; i < averages.size(); ++i)
    {
        if (minimal_CA_distance == -1.0)
        {
            minimal_CA_distance = averages[i][0] - deviations[i][0]
                    * deviation_mult;
        }
        else if (averages[i][0] - (deviations[i][0] * deviation_mult)
                < minimal_CA_distance)
        {
            minimal_CA_distance = averages[i][0] - deviations[i][0]
                    * deviation_mult;
        }
        if (maximal_CA_distance == -1.0)
        {
            maximal_CA_distance = averages[i][0] + deviations[i][0]
                    * deviation_mult;
        }
        else if (averages[i][0] + (deviations[i][0] * deviation_mult)
                > maximal_CA_distance)
        {
            maximal_CA_distance = averages[i][0] + deviations[i][0]
                    * deviation_mult;
        }
    }
}
