/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MOTIF_H_
#define MOTIF_H_

#include <BALL/STRUCTURE/rotamerLibrary.h>

#include <BALL/STRUCTURE/fragmentDB.h>
#include <BALL/STRUCTURE/structureMapper.h>
#include <BALL/STRUCTURE/peptideBuilder.h>
#include <BALL/STRUCTURE/residueRotamerSet.h>
#include <BALL/MATHS/analyticalGeometry.h>
#include <BALL/KERNEL/standardPredicates.h>
#include <BALL/FORMAT/PDBFile.h>
#include <BALL/FORMAT/INIFile.h>
#include <BALL/DATATYPE/stringHashSet.h>


#include "scaffold_selection_utilities.h"


using namespace BALL;
using namespace std;



/**
 * Motif Class. Contains the definition of the enzymatic motif, i.e. most importantly
 * the geometries of the catalytic atoms and the identities of the corresponding residues.
 * Also contains the necessary methods for the calculation of distributions of distances and angles
 * and other properties of the motif geometry.
 */
class Motif
{

private:

		// member variables

	/**
	 * assigns each of the amino acids (except alanine and glycine) 3 sidechain atoms in PDB syntax
	 */
	map<String, vector<String> > catalytic_atoms_map;

	/**
	 * the name and path of the rotamer library file in dunbrack format
	 */
	String rotamer_lib_file;

	/**
	 * the names (have to be in amino acid 3 letter code) of the residues in the motif
	 */
	vector<String> residue_names;

	/**
	 *  the unique user-defined IDs of the residues in the motif
	 */
	vector<String> residue_ids;

	/**
	 * the IDs of the motif residues in a HashSet
	 */
	HashSet<String> residue_id_set;

	/**
	 * minimal possible CA distance of residues in the motif that have a probability higher than
	 * the threshold for at least one of the CA-distributions between the residue pairs
	 */
	double minimal_CA_distance;

	/**
	 * maximal possible CA distance of residues in the motif that have a probability higher than
	 * the threshold for at least one of the CA-distributions between the residue pairs
	 */
	double maximal_CA_distance;

	/**
	 * the minimal probability that a CA distance can have if the residue pair should be considered as a pair hit
	 */
	double probability_threshold;

    /**
     * A motif hit with a backbone clash penalty larger than this
     * value will be skipped
     */
    float maximal_backbone_clash_penalty;

    /**
     * A motif hit with a catalytic atoms geometry penalty larger than this
     * value will be skipped
     */
    float maximal_catalytic_geometry_penalty;

	/**
	 *  number of unique residue pairs in the motif
	 */
	unsigned int pair_number;

	/**
	 *  number of residues in the motif
	 */
	unsigned int residue_number;

	/**
	 *  the desired relative SAS areas (i.e. the percentage of the residue surface that is
	 *  solvent accessible) for the motif residues
	 */
	vector<double> relative_SAS;

	/**
	 * contains a vector for each motif residue with residue objects whose sidechains have the chi angles defined by the rotamers
	 * of the rotamer library
	 */
	vector< vector<Residue> > rotamers;

	/**
	 *  the frequencies of the rotamers, in the same order as in vector #rotamers
	 */
	vector< vector<double> > rotamer_frequencies;

	/**
	 *  the names of the 3 catalytic atoms for each motif residue
	 */
	vector< vector<String> > catalytic_atom_names;

	/**
	 *  the positions of the 3 catalytic atoms for each motif residue
	 */
	vector< vector<Vector3> > catalytic_atom_positions;

	/**
	 *  the positions of the 3 catalytic atoms for each motif residue,
	 *  in one linear vector (more convinient for some tasks, e.g. mapping two atoms sets)
	 */
	vector<Vector3> catalytic_atom_positions_linear;

	/**
	 *  each residue pair of the motif is stored exactly once, as a pair of indices pointing to the
	 * residues of the pair
	 */
	vector< pair<unsigned int, unsigned int> > residue_pairs;

	/**
	 * maps the indices of motif residues to the indices of the corresponding motif residue pairs.
	 * key: the indices i & j of two motif residues i j in the following form: i*10 + j
	 * value: the index of the corresponding motif residue pair (i, j) in vector residue_pairs
	 */
	HashMap<unsigned int, unsigned int> residue_pair_indices;

	/**
	 * in catalytic_atom_distances[i][j][k][l] is the distance of catalytic atom no. k of
	 * motif residue i and catalytic atom  no. l of motif residue j
	 */
	vector< vector< vector< vector<double> > > > catalytic_atom_distances;

	/**
	 * torsion angles for each residue pair and any combination of 2 catalytic atoms of one residue
	 * and 2 catalytic atoms of the other residue
	 * the keys are coded as follows:
	 * 10^0 & 10^1: index of residue 1, 10^2 & 10^3: index of residue 2
	 * 10^4 : atom1 of residue1, 10^5 atom2 of residue 1 10^6 : atom1 of residue2, 10^7 atom2 of residue 2
	 * example: key: 2 1 2 0 01 00 is derived from residue 0, atoms 0 and 2, and residue 1, atoms 1 and 2
	 */
	HashMap<unsigned int, double> catalytic_atom_torsions;

	/**
	 * the frequency-weighted averages of all 4 geometric variables (CA distance, Angles CB1 - CA1 - CA2, CA1 - CA2 - CB2,
	 * torsion CB1 - CA1 - CA2 - CB2) for all residue pairs in the motif
	 */
	vector< vector<double> > averages;

	/**
	 * the standard deviation of all 4 geometric variables (CA distance, Angles CB1 - CA1 - CA2, CA1 - CA2 - CB2,
	 * torsion CB1 - CA1 - CA2 - CB2) for all residue pairs in the motif
	 */
	vector< vector<double> > deviations;

	/**
	 * BALL Expressions (similar to regex) for the interaction centers
	 */
	vector<Expression> interaction_center_expressions;

	/**
	 * Positions of the interaction centers, order as in #interaction_center_expressions
	 */
	vector<Vector3> interaction_center_positions;

	/**
	 * cutoff radii for interaction centers (largest distance of fitting atom allowed from desired position)
	 */
	vector<double> interaction_center_cutoff_radii;

	/**
	 * the largest cutoff radius of all interaction centers
	 */
	double largest_interaction_center_cutoff_radius;

	/**
	 * contains the normal distribution table
	 */
	static float const normal_distribution[];


	/**
	 * the name of the motif
	 */
	String name;

	/**
	 * the pdb system containing the substrate model in position relative to the catayltic atoms
	 */
	System substrate_system;



			///////  private methods ////////////////
	/**
 	 * reads in all parameters of a motif definition file and sets the corresponding member variables.
	 * Also checks if every necessary parameter is present in the file
	 */
	void read_motif_ini_file (String const &file_name);


	/**
	 * creates a map that assigns each amino acid type (no GLY and ALA)
	 * the names of its three catalytic atoms in PDB notation in defined order
	 */
	map<String, vector<String> > create_catalytic_atoms_map ();


	/**
	 * Compute the distribution of the Ca-Cb distances & angles of residues res1 & res2.
	 * The catalytic atom's positions are fixed, all built rotamers are mapped onto the catalytic atoms
	 * Use the rotamers to compute the distributions of the Ca-Ca distance & Ca/Cb angles & torsion.
	 */
	void compute_pairwise_distribution (unsigned int i1, unsigned int i2,
										vector<double>& averages,
										vector<double>& deviations);

	/**
	 * Looks in the distributions of the Ca distances of the motif for the
	 * maximal and the minimal Ca-distance that can theoretically score
	 * a probability higher than zero. Sets members #minimal_CA_distance and #maximal_CA_distance accordingly.
	 */
	void compute_distance_thresholds ();

	/**
	 * computes the distributions (i.e. averages and standard deviations) of the geometric variables (CA-distance, 3 Angles)
	 * for all residue pairs. Addidtionally, the distances of all catalytic atom pairs that are
	 * not in the same residue are computed and stored.
	 * Torsion angles are computed for each combination of 2 catalytic atom positions per 2 residues
	 * (Each torsion is computed for 4 positions. Two of them are derived from one residue, two from the
	 * other. this is done for each residue pair, and for all possible atom pair combinations)
	 */
	void compute_motif_distributions ();


public:

	/////////// Constructors ///////////

	/**
	 *  Default constructor. Yields an empty motif
	 */
	Motif ();

	/**
	 * Constructs a Motif instance form a definition file and
	 * sets all necessary members
	 * */
	Motif (String const &motif_definition_file);

	/**
	 * Destructor
	 */
	virtual ~Motif ();


		////// public methods  ////////////


	/**
	 *  returns name of motif residue with index i
	 */
	const String& get_residue_name (unsigned int i) const;

	/**
	 *  returns ID of motif residue with index i
	 */
	const String& get_residue_id (unsigned int i) const;

	/**
	 * returns a set with all ids of residues in the motif
	 */
	const HashSet<String>& get_residue_id_set() const;

	/**
	 * returns the minimal CA distance of the motif that has a probability higher than the threshold
	 */
	const double get_minimal_CA_distance () const;

	/**
	 *  returns the maximal CA distance of the motif that has a probability higher than the threshold
	 */
	const double get_maximal_CA_distance () const;

	/**
	 *  returns the probability threshold
	 */
	const double get_probability_threshold () const;


    /**
     * returns the maximal backbone clahs penalty
     */
    const float get_maximal_backbone_clash_penalty() const;

    /**
     * returns the maximal catalytic atoms geometry penalty
     */
    const float get_maximal_catalytic_geometry_penalty() const;

	/**
	 *  returns the number of motif residue pairs
	 */
	const unsigned int get_residue_pair_count () const;

	/**
	 *  returns the number of motif residues
	 */
	const unsigned int get_residue_count () const;

	/**
	 *  returns the desired relative SAS area for motif residue i
	 */
	const double get_relative_SAS_for_residue (unsigned int i) const;

	/**
 	 * returns a pointer to built rotamer rot of residue res
 	 * (built rotamer = a residue with the chi torsions of the rotamer rot)
 	 */
	Residue* get_rotamer (unsigned int res, unsigned int rot);

	/**
	 *  returns the frequency of rotamer "rotamer" for motif residue res
	 */
	const double get_rotamer_frequency (unsigned int res, unsigned int rotamer) const;

	/**
	 *  returns the number of different rotamers for motif residue i
	 */
	const unsigned int get_number_of_rotamers_for_residue (unsigned int i) const;

	/**
	 *  returns the name of the catalytic atom 'atom' for motif residue res
	 */
	const String& get_catalytic_atom_name (unsigned int res, unsigned int atom) const;

	/**
	 *  returns the position of the catalytic atom 'atom' for motif residue res
	 */
	const Vector3& get_catalytic_atom_position (unsigned int res, unsigned int atom) const;

	/**
	 * returns the positions of all catalytic atoms in order as defined in the motif file
	 */
	const vector<Vector3>& get_catalytic_atom_positions_vector () const;

	/**
	 *  returns motif residue pair i
	 */
	const pair<unsigned int, unsigned int> get_residue_pair (unsigned int i) const;

	/**
	 * returns the index of the motif residue pair formed by motif residues i & j
	 */
	const unsigned int get_residue_pair_index(unsigned int i, unsigned int j) const;

	/**
	 *  returns the weighted average values for all 4 geometric variables for motif residue pair i
	 */
	const vector<double>& get_averages_for_residue_pair (unsigned int i) const;

	/**
	 *  returns standard deviation values for all 4 geometric variables for motif residue pair i
	 */
	const vector<double>& get_deviations_for_residue_pair (unsigned int i) const;

	/**
	 *  returns the distance of catalytic atom a1 in
	 *  motif residue res1 to catalytic atom a2 in motif residue res2
	 */
	const double get_catalytic_atom_distance (unsigned int res1, unsigned int res2, unsigned int a1, unsigned int a2) const;

	/**
	 * returns the torsion angle of atoms a11 , a12 of residue res1 and atoms a21, a22 of residue res2
	 * where a11 < a12, a21 < a22, res1 < res2
	 */
	const double get_catalytic_atom_torsion (unsigned int res1, unsigned int res2, unsigned int a11, unsigned int a12, unsigned int a21, unsigned int a22) const;

	/**
	 * returns the BALL Expression of interaction center i
	 */
	const Expression get_interaction_center_expression (unsigned int i) const;

	/**
	 * returns the position of interaction center i
	 */
	const Vector3 get_interaction_center_position (unsigned int i) const;

	/**
	 * returns the cutoff radius of interaction center i
	 */
	const double get_interaction_center_cutoff_radius (unsigned int i) const;

	/**
	 * returns the distance between catalytic atom a1 in residue res1 and catalytic atom a2 in res2
	 */
	const double get_maximal_interaction_cutoff_radius () const;

	/**
	 * returns the number of interaction centers
	 */
	const unsigned int get_interaction_center_count () const;

	/**
	 * returns the name of the motif
	 */
	const String get_name () const;

    /**
     * returns the probability of value d in the standard normal distribution
     */
    double get_normal_probability(double d) const;

	/**
	 * returns a pointer to the substrate pdb system;
	 */
	System* get_substrate_system ();
};

#endif /*MOTIF_H_*/
