/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "motifConstruction.h"
using namespace BALL;
using namespace std;

/*
 * Constructor
 * reads in the PDB structure, and maps the motif
 * on the residues at the provided positions in the motif
 */
MotifConstruction::MotifConstruction(Motif& motif_ref, String parameter_file)
{
    motif = motif_ref;
    read_parameter_file(parameter_file);
    PDBFile pfile;
    pfile.open(pdb_file);
    pfile >> system;
    pfile.close();
    anchor_residues.clear();
    unsigned int i = 0;

    // loop over all motif residues
    for (; i < motif.get_residue_count(); ++i)
    {
        // loop over all residues in the current structure

        for (ResidueIterator resit = system.beginResidue(); resit
                != system.endResidue(); ++resit)
        {
            if (resit->isAminoAcid())
            {
                // if the current residue is the current anchor,
                // store a pointer to it
                if (resit->getID() == sequence_positions[i]
                        && resit->getChain()->getName() == chains[i])
                {
                    anchor_residues.push_back(&*resit);
                }
            }
        }
    }

    motif_hit = MotifHit(system.getName(), anchor_residues, motif);
    for (i = 0; i < motif.get_residue_count(); ++i)
    {
        set<unsigned int> rotamer_set;
        for (unsigned int j = 0; j
                < motif.get_number_of_rotamers_for_residue(i); ++j)
        {
            rotamer_set.insert(j);
        }
        motif_hit.insert_allowed_rotamers(i, rotamer_set);
    }
}

MotifConstruction::~MotifConstruction()
{
}

void MotifConstruction::read_parameter_file(String file_name)
{
    INIFile f(file_name);
    f.read();
    String value;

    for (unsigned int i = 1; i <= motif.get_residue_count(); ++i)
    {
        String section = "attachment_position" + String(i);
        value = f.getValue(section, "index");
        if (value.isDigit())
            sequence_positions.push_back(value);
        else
            throw "attachment positions must be integers";
        value = f.getValue(section, "chain");
        chains.push_back(value);
    }

    pdb_file = f.getValue("paths", "scaffold_file");
    output_file = f.getValue("paths", "output_file");
}


void MotifConstruction::graft_motif_on_scaffold()
{
    Expression backbone_expression("backbone()");
    Expression hydrogen_ex("element(H)");
    // graft all rotamers on the backbone positions of the scaffold
    for (unsigned int i = 0; i < motif.get_residue_count(); ++i)
    {
        // get a pointer to the chain of the current anchor residue
        // so we know were to insert the motif residue
        Chain * c = anchor_residues[i]->getChain();

        // loop over all atoms in the motif residue
        AtomIterator ait = motif_hit.get_rotamer(i)->beginAtom();
        for (; ait != motif_hit.get_rotamer(i)->endAtom();)
        {
            // hydrogens are erased from the built motif rotamers,
            // because PDB structures usually do not contain hydrogen positions
            if (hydrogen_ex(*ait))
            {
                // remove the hydrogens from the motif residue
                AtomIterator temp = ait;
                ait++;
                temp->destroy();
            }
            else
            {
                if (backbone_expression(*ait))
                    ait->setPosition(anchor_residues[i]->getAtom(ait->getName())
                                     ->getPosition());
                ++ait;
            }
        }

        // give the motif residue the same id as the anchor residue
        String id = anchor_residues[i]->getID();
        motif_hit.get_rotamer(i)->setID(id);

        // we need to determine if we have to insert the motif
        // residue before or after the anchor (the anchor will later be removed)
        if (anchor_residues[i]->isCTerminal())
        {
            // insert the motif residue into the proper chain next to the
            // anchor residue
            c->insertBefore(*motif_hit.get_rotamer(i), *anchor_residues[i]);
        }
        else
        {
            // insert the motif residue into the proper chain next to
            // the anchor residue
            c->insertAfter(*motif_hit.get_rotamer(i), *anchor_residues[i]);
        }
        // remove the anchor
        c->remove(*anchor_residues[i]);
        // clear the anchor residue of all atoms, just to be sure
        anchor_residues[i]->clear();
    }

    PDBFile out;
    out.open(output_file, ios::out);

    String header = pdb_file.getField(pdb_file.countFields("/") - 1, "/")
            + " WITH MOTIF AT";

    for (unsigned int i = 0; i < sequence_positions.size(); ++i)
    {
        header += " " + sequence_positions[i];
        if (chains.size() > 0)
        {
            header += ":" + chains[i];
        }
    }

    system.setName(header);
    out << system;
    out.close();
}

/*
 * checks if there is an atom that fits the BALL expression of
 * an interaction center closer than the cutoff radius around the
 * reconstructed center position. the center position
 * is reconstructed using the reconstruction matrix
 */
void MotifConstruction::determine_interaction_centers()
{

    Vector3 lower(Limits<float>::max());
    Vector3 upper(Limits<float>::min());

    ResidueIterator res_it = system.beginResidue();
    AtomIterator ait;
    Vector3 position;
    for (; res_it != system.endResidue(); ++res_it)
    {
        if (res_it->isAminoAcid())
        {
            ait = res_it->beginAtom();
            for (; ait != res_it->endAtom(); ++ait)
            {
                position = ait->getPosition();

                // look for the smallest and largest values of the
                // x,y,z coordinates of all included residue atoms to
                // determine the positions of the corners of the HashGrid
                if (position.x < lower.x)
                    lower.x = position.x;
                if (position.y < lower.y)
                    lower.y = position.y;
                if (position.z < lower.z)
                    lower.z = position.z;
                if (position.x > upper.x)
                    upper.x = position.x;
                if (position.y > upper.y)
                    upper.y = position.y;
                if (position.z > upper.z)
                    upper.z = position.z;
            }
        }
    }

    // create a Hashgrid that will hold all potential interaction center atoms
    HashGrid3<Atom*> interaction_atoms_grid(
           lower - Vector3(motif.get_maximal_interaction_cutoff_radius()),
           upper - lower
                + Vector3( 2.0 * motif.get_maximal_interaction_cutoff_radius()),
           motif.get_maximal_interaction_cutoff_radius());

    unsigned int i;

    // loop over all atoms of the protein
    for (res_it = system.beginResidue(); res_it != system.endResidue(); ++res_it)
    {
        if (res_it->isAminoAcid())
        {
            ait = res_it->beginAtom();
            for (; ait != res_it->endAtom(); ++ait)
            {
                // loop over all interaction centers
                for (i = 0; i < motif.get_interaction_center_count(); ++i)
                {
                    // if the atom fits one of the interaction center BALL
                    // expressions, insert it into the grid
                    if (motif.get_interaction_center_expression(i)(*ait))
                    {
                        interaction_atoms_grid.insert(ait->getPosition(), &*ait);
                    }
                }
            }
        }
    }

    motif_hit.check_interaction_centers(interaction_atoms_grid, true);
}

/*
 * recontructs the positions of all spheres for each motif hit,
 * and checks of there are any backbone atoms inside the spheres.
 * if so, computes a penalty.
 */
void MotifConstruction::construct_ligand_position()
{
    System* ligand_system = motif.get_substrate_system();
    StructureMapper mapper;
    mapper.setTransformation(motif_hit.get_mapping_matrix());
    ligand_system->apply(mapper);
    System ls = *ligand_system;
    system.spliceAfter(ls);
}

/*
 * identifies the anchor residues for the motif hit and
 * maps the motif residues onto these anchors.
 * writes the altered structure into a PDB file
 */
void MotifConstruction::construct_motif()
{
    Vector3 lower(Limits<float>::max());
    Vector3 upper(Limits<float>::min());
    Vector3 position;

    unsigned int minimal_chain_length = 20;
    AtomIterator ait;
    for (ResidueIterator res_it = system.beginResidue(); res_it
            != system.endResidue(); ++res_it)
    {
        if (res_it->isAminoAcid() && res_it->getAtom("CA")
                && (res_it->getChain()->countResidues() > minimal_chain_length))
        {
            ait = res_it->beginAtom();
            for (; ait != res_it->endAtom(); ++ait)
            {
                position = ait->getPosition();

                // look for the smallest and largest values of the
                // x,y,z coordinates of all
                // included residues to determine the positions
                // of the corners of the HashGrid
                if (position.x < lower.x)
                    lower.x = position.x;
                if (position.y < lower.y)
                    lower.y = position.y;
                if (position.z < lower.z)
                    lower.z = position.z;
                if (position.x > upper.x)
                    upper.x = position.x;
                if (position.y > upper.y)
                    upper.y = position.y;
                if (position.z > upper.z)
                    upper.z = position.z;
            }
        }
    }

    // move the corners of the hashgrid one boxlength an both sides,
    // to be sure that every residue
    // fits into the grid
    lower.x -= 3.7;
    lower.y -= 3.7;
    lower.z -= 3.7;
    upper.x += 3.7;
    upper.y += 3.7;
    upper.z += 3.7;

    // Hashgrid with all backbone atoms for clash checking
    // spacing is the sum of the 2 largest vdW-radii (Sulfur)
    HashGrid3<Atom*> backbone_atoms_grid = HashGrid3<Atom*> (lower
            - Vector3(0.1F), upper - lower + Vector3(0.2F), 3.7F);

    // this is used to determine which atoms should be checked for clashes
    Expression backbone_expression("backbone() OR name(CB)");

    // loop over all atoms of the protein
    for (ResidueIterator res_it = system.beginResidue(); res_it
            != system.endResidue(); ++res_it)
    {
        if (res_it->isAminoAcid() && res_it->getAtom("CA")
                && (res_it->getChain()->countResidues() > minimal_chain_length))
        {
            ait = res_it->beginAtom();
            for (; ait != res_it->endAtom(); ++ait)
            {
                // if the atom is a backbone atom or a CB atom,
                // insert it into the grid
                if (backbone_expression(*ait))
                {
                    backbone_atoms_grid.insert(ait->getPosition(), &*ait);
                }
            }
        }
    }

    unsigned int i;

    //construct the motif onto the anchors
    motif_hit.construct_motif_rotamers(
                                   backbone_atoms_grid,
                                   motif.get_maximal_backbone_clash_penalty());
    cout << "catalytic geometry penalty score: "
         << motif_hit.get_catalytic_atoms_geometry_penalty()
            << endl;
    motif_hit.compute_mapping_matrix();

    if (motif.get_interaction_center_count() > 0)
    {
        determine_interaction_centers();
        for (i = 0; i < motif.get_interaction_center_count(); ++i)
        {
            const Atom* ia = motif_hit.get_interaction_atom(i);
            if (ia != NULL)
            {
                cout << "Interaction center " << i << ": " << "Atom "
                        << ia->getName() << " of residue "
                        << ia->getResidue()->getName() << " "
                        << ia->getResidue()->getID() << " in chain "
                        << ia->getChain()->getName() << endl;
            }
            else
            {
                cout << "No compatible atom found for interaction center " << i
                        << endl;
            }
        }
    }
    if (motif.get_substrate_system() != NULL)
    {
        cout << "Determining substrate position..." << endl;
        construct_ligand_position();
    }
    graft_motif_on_scaffold();
    cout << "Motif successfully grafted onto scaffold result in PDB file "
            << output_file << endl;
}
