/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MOTIFHIT_H_
#define MOTIFHIT_H_

#include <BALL/MATHS/matrix44.h>
#include <BALL/KERNEL/system.h>
#include <BALL/COMMON/limits.h>
#include <BALL/STRUCTURE/RMSDMinimizer.h>
#include <BALL/DATATYPE/hashGrid.h>
#include <BALL/KERNEL/PTE.h>
#include <vector>
#include <set>
#include "motif.h"


using namespace std;
using namespace BALL;

/**
 *  holds all necessary information of a motif hit
 * and contains methods to compute the characterizing penalty scores for this hit
 */
class MotifHit
{
public:

	/**
	 * Constructor,
	 * needs the name of the PDB system, the vector of valid residues and the motif as arguments
	 */
	MotifHit(String const &system_name,
	         vector<Residue*> const &res,
	         Motif& mot);

	MotifHit();

	/**
	 * Destructor
	 */
	virtual ~MotifHit();

	/**
	 * checks each combination of rotamers of the motif residues constructed on the backbone of the
	 * target protein at the positions indicated by the motif hit. the combination that minimizes the catalytic atoms
	 * penalty score will be used as the reconstructed motif hit.
	 * the rotamers are checked before the actual search if they collide with the backbone. if rotamer clashes too
	 * strong with the scaffold backbone, it is not used in the following search.
	 */
	bool construct_motif_rotamers (HashGrid3<Atom*> const &backbone_atoms_grid, double backbone_clash_threshold);

	/**
	 * Computes for a transformation matrix that maps the coordinates of the motif catalytic atoms
	 * onto the coordinates of these atoms in the reconstructed motif (according to the hit)
	 */
	void compute_mapping_matrix ();

	/**
	 * checks if there is an atom that fits the BALL expression of an interaction center
	 * closer than the cutoff radius around the reconstructed center position. the center position
	 * is recontructed using the reconstruction matrix
	 */
	bool check_interaction_centers (HashGrid3<Atom*> const &interaction_atoms_grid, bool force_search_all);

	/**
	 * Checks if the sidechain atoms of the constructed motif residues
	 * clash with any backbone atoms of the protein.
	 * clashes are penalized according to the severity of the clash
	 * ( two atoms clash if their distance is smaller than their vdW-radii.) the penalty
	 * depends on how much smaller the distance is
	 */
	void check_backbone_clashes (HashGrid3<Atom*> const &backbone_atoms_grid);

	/**
	 * recontructs the positions of all spheres for the motif hit,
	 * and checks of there are any backbone atoms inside the spheres.
	 * if so, computes a penalty.
	 */
	void check_ligand_clashes(HashGrid3<Atom*> const &backbone_atoms_grid);

	/**
	 * checks if the relative SAS areas of the anchor residues
	 * are similar to the SAS wanted for the motif residues,
	 * and compute a penalty according to the discrepancy
	 */
	bool compute_relative_SAS_score (vector<double> const &residue_SAS_areas);

	/**
	 * set the residue index at position pos to value i
	 */
	void set_residue_index(unsigned int pos, unsigned int i);

	/**
	 * add a set of allowed rotamers for residue i, the iterators mark the beginning and the end of the set to add
	 */
	void insert_allowed_rotamers(unsigned int i, const set<unsigned int>&);

	/**
	 * returns the rotamer probability score of this motif hit
	 */
	const double get_rotamer_probability() const;

	/**
	 * returns the catalytic atoms geometry penalty score of this motif hit
	 */
	const double get_catalytic_atoms_geometry_penalty() const;

	/**
	 * returns the relative SASA penalty score of this motif hit
	 */
	const double get_relative_SAS_penalty() const;

	/**
	 * returns the backbone clash penalty score of this motif hit
	 */
	const double get_backbone_clash_penalty() const;

	/**
	 * returns the spheres clash penalty score of this motif hit
	 */
	const double get_ligand_clash_penalty() const;

	/**
	 * returns the interaction center penalty score of this motif hit
	 */
	const double get_interaction_centers_penalty() const;

	/**
	 * returns a unique identifier String for the motif hit. it consists of the name of the PDB structure,
	 * the ID and Chain ID of each anchor residue of the motif hit in the same order as the residues in the motif
	 * definition file / motif object
	 */
	const String get_ID() const;

	/**
	 *
	 * @return the indices of the identified rotamers in the rotamer library
	 *         list separated by spaces
	 */
	const String get_rotamer_indices() const;


	/**
	 * returns a pointer to the Atom that serves as the interaction center i
	 * if no compatible Atom was found, returns a NULL pointer
	 */
	const Atom* get_interaction_atom (unsigned int i) const;

	/**
	 * returns the matrix that transforms the coordinate system of the motif definition
	 * to the adequate coordinates for the motif hit
	 */
	Matrix4x4 get_mapping_matrix() const;

	/**
	 * returns a pointer to the Rotamer determined for motif residue no. i attachment
	 */
	Residue* get_rotamer(unsigned int i);

	/**
	 * sets the partial geometric penalty score for rotamer pair rot1, rot2 for pair pair_index
	 */
	void set_pair_rotamer_score(unsigned int pair_index, unsigned int rot1, unsigned int rot2, double score);

	bool is_wt_hit;

private:

	/**
	 * for each motif residue pair, maps the partial pairwise geometric penalty score of rotamer pair i,j to key i * 10000 + j
	 */
	vector<HashMap<unsigned int, double> > pair_rotamer_scores;

	/**
	 * the indices of rotamers that did lead to good catalytic atoms geometries at the pair hit determination stage
	 * of MotifSearch
	 */
	vector<set<unsigned int> > allowed_rotamer_sets;


	/**
	 * partial_rotamer_pair_score[i] contains the rotamer pair scores for motif residue pair i
	 * The HashMap keys are combinations of the rotamer indices in the motifs rotamer sets, the value is the
	 * partial catalytic geometry penalty score
	 */
//	vector<HashMap<unsigned int, double> > partial_rotamer_pair_scores;

	/**
	 * the indices in vector all_residues of the residues where the motif can be attached,
	 * in the same order as the residues are stored in the motif object
	 */
	vector<unsigned int> indices;

	/**
	 * the indices of the motif object's rotamers that reproduce the motif geometry best
	 */
	vector<unsigned int> rotamer_indices;

	/**
	 * the product of the frequencies of the rotamers of this motif hit
	 */
	double rotamer_probability;

	/**
	 * the rmsd differences of catalytic atom distances of the motif hit (i.e. the motif residues constructed on the protein
	 * backbone with at the positions indicated by the hit with the rotamers identified in rotamer_indices)
	 * and the original catalytic atom distances in the original motif
	 */
	double catalytic_atoms_geometry_penalty;

	/**
	 * the rmsd of differences in relative SAS areas between the desired rel SASAs defined in
	 * the motif object and the SASAs of the residues at the motif hit positions
	 */
	double relative_SAS_penalty;

	/**
	 * the sum of the squared violation of minimum distances between the sidechain atoms
	 * of the costructed motif on the positions indicated by indices and the backbone atoms
	 * and CB of the potential scaffold
	 */
	double backbone_clash_penalty;

	/**
	 * the sumligand of the squared violation of minimum distances between atoms
	 * of the ligand and the backbone atoms and CB of the potential scaffold
	 */
	double ligand_clash_penalty;

	/**
	 * the sum of squares of the distance of each interaction center atom
	 * to the ideal recontructed position of the interaction center
	 */
	double interaction_centers_penalty;

	/**
	 * the matrix that maps the motif catalytic atom coordinates onto the
	 * coordinates of the atoms in a recosntructed motif hit
	 */
	Matrix4x4 mapping_matrix;

	/**
	 * contains the vdW-Radius of an oxygen atom
	 */
	static const double oxygen_radius = 1.52;

	/**
	 * contains the vdW-Radius of a carbon atom
	 */
	static const double carbon_radius = 1.7;

	/**
	 * contains the vdW-Radius of a nitrogen atom
	 */
	static const double nitrogen_radius = 1.55;

	/**
	 * contains the vdW-Radius of a sulfur atom
	 */
	static const double sulfur_radius = 1.85;

	/**
	 * Name of the pdb_file the motif hit is located in
	 */
	String pdb_name;

	/**
	 * contains pointers to all residues in the current protein that are amino acids
	 */
	vector<Residue*> all_residues;

	/**
	 * contains pointers to the atoms serving as interaction centers
	 * if no atom could be found, a NULL pointer is included at the appropriate position
	 */
	vector<Atom*> interaction_atoms;

	/**
	 * motif object defining the geometry, rel. SAS areas, backbone-atome free spheres and
	 * interacting atoms to look for
	 */
	Motif motif;

	/**
	 * recursively constructs every rotamer combination on the positions of a
	 * motif hit, and identifies the rotamer set that has the most similar
	 * geometry to the original motif.
	 * similarity is measured by the difference of the distances between the
	 * catalytic atoms of the motif residues/rotamers and the difference of
	 * torsion angles of any 2 catayltic atoms of any 2 residues between the motif
	 * residues and the built rotamers
	 */
	void rotamer_construction_recursion (vector<unsigned int>& motif_rotamer_set,
		  		   					     vector<unsigned int>& minimal_rotamer_set,
								   		 double& minimal_geometry_penalty,
								   		 double partial_score);



};

#endif /*MOTIFHIT_H_*/
