/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "motifSearch.h"
using namespace BALL;
using namespace std;

/*
 * this constructor assigns its arguments to the respective
 * member variables of the Motifsearch
 */
MotifSearch::MotifSearch(System& pdb_system, Motif& motif_ref,
                         Parameters const &parameters_ref) :
    motif(motif_ref), system(pdb_system)

{
    parameters = parameters_ref;
    lower = Vector3(Limits<float>::max());
    upper = Vector3(Limits<float>::min());
}

MotifSearch::~MotifSearch()
{
}

/*
 * compute the probabilities for each geometric value in vector geom_values
 * (i.e. Ca- distance and angles phi, psi, tau).
 * probability here means similarity of the residue pair that led to geom_values
 * to a residue pair in the motif (using the average values
 * and the standard deviations of the geometric values in the motif)
 */
vector<double> MotifSearch::get_probabilities(vector<double> const &geom_values)
{
    vector<double> ret;

    // loop over the distributions of all the residue pairs of the motif
    for (unsigned int i = 0; i < motif.get_residue_pair_count(); ++i)
    {
        double global_score = 0.0;
        // cycle over Ca- distance and angles phi, psi, tau
        for (int j = 0; j < 4; ++j)
        {
            // Z-conversion (normalize to standard normal distribution) , so we can use the probability table
            double z = (geom_values[j]
                    - motif.get_averages_for_residue_pair(i)[j])
                    / (motif.get_deviations_for_residue_pair(i)[j]);

            double prob;
            // calculate probability,i.e. how similar the geometric value of the observed
            // residue pair is to the one in the motif
            if (z <= 3.49 && z >= -3.49)
            {
                prob = 1.0 - 2 * (motif.get_normal_probability(z));
            }
            // if the probability of one geometric value is zero, the global probability for this hit
            // is also zero, and we can stop the loop
            else
            {
                global_score = 0.0;
                break;
            }
            // the global probability/similarity is the product ((sum of the logarithms)) of the
            // 4 single probabilies for Ca- distance and angles phi, psi, tau
            global_score += log(prob);
        }
        // store the probability score
        ret.push_back(global_score);
    }
    // returns a vector with a probability score for each motif pair
    return ret;
}

HashSet<Residue*> MotifSearch::get_secondary_structure_residues()
{
    ChainIterator cit = system.beginChain();
    for (; cit != system.endChain(); ++cit)
    {
        SecondaryStructureProcessor sspro;
        cit->apply(sspro);
    }
    HashSet<Residue*> ss_core_residues;

    unsigned int count = 0;

    ResidueIterator it;
    ResidueReverseIterator rit;
    SecondaryStructureIterator sit = system.beginSecondaryStructure();
    for (; sit != system.endSecondaryStructure(); ++sit)
    {
        SecondaryStructure::Type sst = sit->getType();
        if (sst == SecondaryStructure::HELIX && sit->countResidues() > 2
                * parameters.get_helix_boundary_length())
        {
            if (parameters.verbose)
                cout << "Helix: ";
            rit = sit->rbeginResidue();
            for (count = 0; count < parameters.get_helix_boundary_length() - 1; ++count)
            {
                ++rit;
            }
            count = 0;
            it = sit->beginResidue();
            for (; it != sit->endResidue(); ++it)
            {
                if (count < parameters.get_helix_boundary_length())
                {
                    count++;
                }
                else if (*rit == *it)
                {
                    break;
                }
                else
                {
                    ss_core_residues.insert(&*it);
                    if (parameters.verbose)
                        cout << it->getID() << ":" << it->getChain()->getName()
                                << ", ";
                }
            }
            if (parameters.verbose)
                cout << endl;
        }
        else if (sst == SecondaryStructure::STRAND && sit->countResidues() > 2
                * parameters.get_strand_boundary_length())
        {
            if (parameters.verbose)
                cout << "Strand: ";
            rit = sit->rbeginResidue();
            for (count = 0; count < parameters.get_strand_boundary_length() - 1; ++count)
            {
                ++rit;
            }
            count = 0;
            it = sit->beginResidue();
            for (; it != sit->endResidue(); ++it)
            {
                if (count < parameters.get_strand_boundary_length())
                {
                    count++;
                }
                else if (*rit == *it)
                {
                    break;
                }
                else
                {
                    ss_core_residues.insert(&*it);
                    if (parameters.verbose)
                        cout << it->getID() << ":" << it->getChain()->getName()
                                << ", ";
                }
            }
            if (parameters.verbose)
                cout << endl;
        }
    }
    return ss_core_residues;

}

/*
 * Calculate a vector of residue pairs whose Ca distance is
 * smaller than the value of the max_distance variable
 * and larger than the value of the min_distance variable
 */
void MotifSearch::get_residue_pairs(
                                    HashSet<Residue*> const &pocket_residues,
                                    HashSet<Residue*> const &secondary_structure_residues)
{
    // clear the containers, so we can start from scratch with computing the residue pairs
    vector<double> residue_SAS_areas;
    all_residues.clear();
    pair_vector.clear();

    // Position vector
    Vector3 position;

    // pointers to all residues that are amino acids and contain a CA atom
    // are stored. Other residues are removed from the protein
    AtomIterator atom_it;
    HashSet<Residue*>::const_iterator p_it;
    if (parameters.verbose)
        cout << "Considering residue positions ";
    for (ResidueIterator res_it = system.beginResidue(); res_it
            != system.endResidue(); ++res_it)
    {
        if (res_it->getAtom("CA"))
        {
            position = res_it->getAtom("CA")->getPosition();
            // look for the smallest and largest values of the x,y,z coordinates of all
            // included residues to determine the positions of the corners of the HashGrid
            if (position.x < lower.x)
                lower.x = position.x;
            if (position.y < lower.y)
                lower.y = position.y;
            if (position.z < lower.z)
                lower.z = position.z;
            if (position.x > upper.x)
                upper.x = position.x;
            if (position.y > upper.y)
                upper.y = position.y;
            if (position.z > upper.z)
                upper.z = position.z;
        }
        if (res_it->isAminoAcid() && res_it->getAtom("CA")
                && (res_it->getAtom("CB") || res_it->getName() == "GLY")
                && (res_it->getChain()->countResidues()
                        > parameters.get_minimal_chain_length())
                && (!parameters.get_compute_surface_pockets()
                        || pocket_residues.has(&*res_it))
                && !secondary_structure_residues.has(&*res_it))
        {
            // store a pointer to the residue, it will be included in the motif search
            all_residues.push_back(&*res_it);
            if (parameters.verbose)
                cout << res_it->getID() << ":" << res_it->getChain()->getName()
                        << ", ";
        }
    }
    if (parameters.verbose)
        cout << endl;

    double maximal_CA_distance = motif.get_maximal_CA_distance();
    double minimal_CA_distance = motif.get_minimal_CA_distance();

    // move the corners of the hashgrid one boxlength an both sides, to be sure that every residue
    // fits into the grid
    lower.x -= maximal_CA_distance;
    lower.y -= maximal_CA_distance;
    lower.z -= maximal_CA_distance;
    upper.x += maximal_CA_distance;
    upper.y += maximal_CA_distance;
    upper.z += maximal_CA_distance;

    // squared distances
    double squared_max_distance = maximal_CA_distance * maximal_CA_distance;
    double squared_min_distance = minimal_CA_distance * minimal_CA_distance;
    double distance;

    // Check what kind of algorithm should be used for calculating the
    // residue pairs
    bool brute_force = false;
    float memory = SysInfo::getAvailableMemory();
    if (memory != -1)
    {
        memory *= 0.7;
        float
                min_spacing =
                        HashGrid3<const unsigned int>::calculateMinSpacing(
                                                                           (LongSize)memory,
                                                                           upper
                                                                                   - lower
                                                                                   + Vector3(
                                                                                             0.2F));

        // if there isn't enough memory space for the grid, use the brute force approach
        if (min_spacing > maximal_CA_distance)
        {
            brute_force = true;
        }
    }

    // if there isn't enaough memory, use an algorithm that needs less memory, but is slower
    // Brute force algorithm (O(n^2))
    if (brute_force)
    {
        // loop over all residues of the system
        for (unsigned int i1 = 0; i1 < all_residues.size(); ++i1)
        {
            position = all_residues[i1]->getAtom("CA")->getPosition();
            unsigned int i2 = i1 + 1;

            // loop over all residues of the system that weren't used in the outer loop before
            for (; i2 < all_residues.size(); ++i2)
            {
                // if the Ca distance of the two residues is in the allowed interval, add them to pair_vector
                distance
                        = position.getSquareDistance(
                                                     all_residues[i2]->getAtom(
                                                                               "CA")->getPosition());
                if (distance < squared_max_distance && distance
                        > squared_min_distance)
                {
                    pair_vector.push_back(pair<unsigned int, unsigned int> (i1,
                                                                            i2));
                }
            }
        }
    }
    // There is enough memory to build a Hash grid for faster determination of the motif pairs
    else
    {
        // Algorithm using a 3D hash grid
        //
        // Use a hash grid with box length "max_distance" to determine all
        // residue pairs with Ca-distance in the interval [min_distance, max_distance].
        // these will be tested later, if they are geometrically similar to a residue pair in the motif

        // create the Hashgrid
        // enlarge the box by some constant to be sure not to run into
        // numerical problems
        // 					     	   origin                      size 					    spacing
        HashGrid3<unsigned int> grid(lower - Vector3(0.1F), upper - lower
                + Vector3(0.2F), maximal_CA_distance + 0.1F);

        unsigned int residue_index = 0;

        // insert all residues into the grid
        for (vector<Residue*>::iterator res_it = all_residues.begin(); res_it
                != all_residues.end(); ++res_it)
        {
            // insert residue into hash grid at the position of its Ca-atom
            grid.insert((*res_it)->getAtom("CA")->getPosition(), residue_index);
            ++residue_index;
        }

        // iterate over all boxes of the hash grid
        HashGrid3<unsigned int>::BoxIterator box_it = grid.beginBox();
        for (; box_it != grid.endBox(); ++box_it)
        {
            Position x, y, z;
            grid.getIndices(*box_it, x, y, z);

            // look in the current box for residue pairs with Ca atoms in the allowed range
            HashGridBox3<unsigned int>::ConstDataIterator bit =
                    box_it->beginData();
            for (; bit != box_it->endData(); ++bit)
            {
                const Vector3& bit_pos =
                        all_residues[*bit]->getAtom("CA")->getPosition();

                // iterate over all residues after bit
                HashGridBox3<unsigned int>::ConstDataIterator tit = bit;
                ++tit;
                for (; tit != box_it->endData(); ++tit)
                {
                    distance
                            = bit_pos.getSquareDistance(
                                                        all_residues[*tit]->getAtom(
                                                                                    "CA")->getPosition());

                    // if Ca distance is in allowed distance range, add the residue pair to pair_vector
                    if (distance < squared_max_distance && distance
                            > squared_min_distance)
                    {
                        pair_vector.push_back(
                                              pair<unsigned int, unsigned int> (
                                                                                *bit,
                                                                                *tit));
                    }
                }
            }
            // iterator over the neighbour boxes
            for (Index xi = -1; xi <= 1; xi++)
            {
                const Position nx = x + xi;
                for (Index yi = -1; yi <= 1; yi++)
                {
                    const Position ny = y + yi;
                    for (Index zi = -1; zi <= 1; zi++)
                    {
                        // central box was treated separately before
                        if ((xi == 0) && (yi == 0) && (zi == 0))
                            continue;

                        HashGridBox3<unsigned int>* neighbour_box =
                                grid.getBox(nx, ny, z + zi);

                        // if the neighbour box is empty, or the box has
                        // no neighbour at this position (i.e. it is situated at a side of the hash grid),
                        // or the neighbour_box is "smaller", i.e. it has been arleady checked againtst the central box
                        // when itself was the central box, continue with the next neihbor box
                        // ( the smaller operator also checks for 0 !)
                        if (neighbour_box < &*box_it
                                || neighbour_box->isEmpty())
                        {
                            continue;
                        }

                        // iterate over all residues of current central box
                        HashGridBox3<unsigned int>::ConstDataIterator tit =
                                box_it->beginData();
                        for (; tit != box_it->endData(); tit++)
                        {
                            const Vector3
                                    & atom_pos =
                                            all_residues[*tit]->getAtom("CA")->getPosition();

                            // iterate over all residues of the neighbour box
                            HashGridBox3<unsigned int>::ConstDataIterator bit =
                                    neighbour_box->beginData();
                            for (; bit != box_it->endData(); bit++)
                            {
                                // if the Ca distance is in the allowed interval, add the residue indexes to pair_vector
                                distance
                                        = all_residues[*bit]->getAtom("CA")->getPosition().getSquareDistance(
                                                                                                             atom_pos);
                                if (distance < squared_max_distance && distance
                                        > squared_min_distance)
                                {
                                    pair_vector.push_back(pair<unsigned int,
                                            unsigned int> (*bit, *tit));
                                }
                            }
                        }
                    } // zi
                } // yi
            } // xi
        } // for all boxes
    }
}

/*
 * compute all residue pairs in the structure whose CA/CB geometries are geometrically similar
 * (with a probability higher than a threshold) of the residue pairs in the motif
 */
void MotifSearch::compute_pair_hits()
{
    double log_threshold = log(motif.get_probability_threshold());

    pair_hit_probabilities.clear();
    pair_hits.clear();
    pair_hit_rotamer_scores.clear();

    // add an empty vector for each motif residue pair
    // to the data structures that will hold the hits for this pair and their probability scores
    for (unsigned int temp = 0; temp < motif.get_residue_pair_count(); ++temp)
    {
        vector<pair<unsigned int, unsigned int> > empty_vec;
        pair_hits.push_back(empty_vec);
        vector<double> empty_prob_vec;
        pair_hit_probabilities.push_back(empty_prob_vec);
        vector<vector<pair<unsigned int, unsigned int> > > empty_rot_vec;
        pair_hit_rotamers.push_back(empty_rot_vec);
        vector<vector<double> > empty_rot_score_vec;
        pair_hit_rotamer_scores.push_back(empty_rot_score_vec);
    }

    vector<vector<double> > geom_values, geom_values_swapped;

    unsigned int removed_count = 0, total_count = 0;

    HashSet<unsigned int> full_set;
    for (vector<pair<unsigned int, unsigned int> >::iterator pit =
            pair_vector.begin(); pit != pair_vector.end(); ++pit)
    {
        full_set.insert(pit->first);
        full_set.insert(pit->second);
    }
    vector<HashSet<unsigned int> > allowed_sites(motif.get_residue_count(),
                                                 full_set);

    for (unsigned int j = 0; j < motif.get_residue_pair_count(); ++j)
    {
        HashSet<unsigned int> allowed_set1, allowed_set2;

        // loop over all residue pairs in the current structure whose CA-distance is in the relevant interval
        for (unsigned int i = 0; i < pair_vector.size(); ++i)
        {
            // check if there is a CB atom (if the Residue is not GLY), so we can compute the angles
            if (allowed_sites[motif.get_residue_pair(j).first].has(
                                                                   pair_vector[i].first)
                    && allowed_sites[motif.get_residue_pair(j).second].has(
                                                                           pair_vector[i].second))
            {
                String id1, id2;
                if (parameters.verbose)
                {
                    id1 = (all_residues[pair_vector[i].first])->getID();
                    id2 = (all_residues[pair_vector[i].second])->getID();
                }
                // if this is the first time through the loop, compute the geometric values
                if (j == 0)
                {
                    Vector3 CA1, CA2, CB1, CB2;
                    CA1
                            = all_residues[pair_vector[i].first]->getAtom("CA")->getPosition();
                    if (all_residues[pair_vector[i].first]->getName() == "GLY")
                    {
                        CB1
                                = scsel_util::get_ideal_cb_position_for_gly(
                                                                all_residues[pair_vector[i].first]->getAtom(
                                                                                                            "N")->getPosition(),
                                                                all_residues[pair_vector[i].first]->getAtom(
                                                                                                            "C")->getPosition(),
                                                                CA1);
                    }
                    else
                        CB1
                                = all_residues[pair_vector[i].first]->getAtom(
                                                                              "CB")->getPosition();

                    CA2
                            = all_residues[pair_vector[i].second]->getAtom("CA")->getPosition();
                    if (all_residues[pair_vector[i].second]->getName() == "GLY")
                    {
                        CB2
                                = scsel_util::get_ideal_cb_position_for_gly(
                                                                all_residues[pair_vector[i].second]->getAtom(
                                                                                                             "N")->getPosition(),
                                                                all_residues[pair_vector[i].second]->getAtom(
                                                                                                             "C")->getPosition(),
                                                                CA2);
                    }
                    else
                        CB2
                                = all_residues[pair_vector[i].second]->getAtom(
                                                                               "CB")->getPosition();

                    vector<double> tmp_geom_vals =
                        scsel_util::compute_geometric_values(CA1, CB1, CA2, CB2);

                    geom_values.push_back(tmp_geom_vals);
                    // swap the two values for the two bond angles psi and phi (Cb1-Ca1-Ca2, Ca1-Ca2-Cb2),
                    // so creating a vector that contains the geometric values for the residue pair in swapped order
                    double tmp = tmp_geom_vals[1];
                    tmp_geom_vals[1] = tmp_geom_vals[2];
                    tmp_geom_vals[2] = tmp;
                    geom_values_swapped.push_back(tmp_geom_vals);
                }

                // compute the probability scores for the set of geometric values
                // probability score here means the similarity of the geometric values (Ca-dist, Angles)
                // of the two residues of the current pdb-structure to the geom. values of a residue pair in the motif
                vector<double> probabilities =
                        get_probabilities(geom_values[i]);

                if (parameters.verbose)
                {
                    if ((motif.get_residue_id_set().has(id1)
                            and motif.get_residue_id_set().has(id2))
                            and (id1
                                    == motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   j).first))
                            and (id2
                                    == motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   j).second)))
                    {
                        cout << "Considering pair " << id1
                                << (all_residues[pair_vector[i].first])->getChain()->getName()
                                << " " << id2
                                << (all_residues[pair_vector[i].second])->getChain()->getName()
                                << " for "
                                << motif.get_residue_id(
                                                        motif.get_residue_pair(
                                                                               j).first)
                                << " "
                                << motif.get_residue_id(
                                                        motif.get_residue_pair(
                                                                               j).second)
                                << ": probability is " << ((probabilities[j]
                                == 0.0) ? 0.0 : exp(probabilities[j])) << ". ";
                    }
                }

                // check if the probability score is higher than the threshold,
                // do this for the probability score for each motif pair
                if (probabilities[j] != 0.0 && probabilities[j] > log_threshold)
                {
                    total_count++;
                    bool has_suitable_rotamers = determine_pair_hit_rotamers(i,
                                                                             j);
                    if (has_suitable_rotamers)
                    {
                        pair_hit_probabilities[j].push_back(probabilities[j]);
                        allowed_set1.insert(pair_vector[i].first);
                        allowed_set2.insert(pair_vector[i].second);
                        if (parameters.verbose)
                        {
                            if ((motif.get_residue_id_set().has(id1)
                                    and motif.get_residue_id_set().has(id2))
                                    and (id1
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).first))
                                    and (id2
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).second)))

                                cout << "Adding pair " << endl;
                        }
                    }
                    else
                    {
                        if (parameters.verbose)
                        {
                            if ((motif.get_residue_id_set().has(id1)
                                    and motif.get_residue_id_set().has(id2))
                                    and (id1
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).first))
                                    and (id2
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).second)))

                                cout << "Removing pair: no suitable rotamers "
                                        << endl;
                        }
                        removed_count++;
                    }
                }
                else
                {
                    if (parameters.verbose)
                    {
                        if ((motif.get_residue_id_set().has(id1)
                                and motif.get_residue_id_set().has(id2))
                                and (id1
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       j).first))
                                and (id2
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       j).second)))
                            cout << "Probability too low." << endl;

                    }
                }
            }
            if (allowed_sites[motif.get_residue_pair(j).first].has(
                                                                   pair_vector[i].second)
                    && allowed_sites[motif.get_residue_pair(j).second].has(
                                                                           pair_vector[i].first))
            {
                String id1, id2;
                if (parameters.verbose)
                {
                    id1 = (all_residues[pair_vector[i].second])->getID();
                    id2 = (all_residues[pair_vector[i].first])->getID();
                }

                vector<double> probabilities_swapped =
                        get_probabilities(geom_values_swapped[i]);
                if (parameters.verbose)
                {
                    if ((motif.get_residue_id_set().has(id1)
                            and motif.get_residue_id_set().has(id2))
                            and (id1
                                    == motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   j).first))
                            and (id2
                                    == motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   j).second)))

                        cout << "Considering pair " << id1
                                << (all_residues[pair_vector[i].second])->getChain()->getName()
                                << " " << id2
                                << (all_residues[pair_vector[i].first])->getChain()->getName()
                                << " for "
                                << motif.get_residue_id(
                                                        motif.get_residue_pair(
                                                                               j).first)
                                << " "
                                << motif.get_residue_id(
                                                        motif.get_residue_pair(
                                                                               j).second)
                                << ": probability is "
                                << ((probabilities_swapped[j] == 0.0) ? 0.0
                                                                      : exp(
                                                                            probabilities_swapped[j]))
                                << ". ";
                }

                if (probabilities_swapped[j] != 0.0 && probabilities_swapped[j]
                        > log_threshold)
                {
                    total_count++;
                    bool has_suitable_rotamers =
                            determine_pair_hit_rotamers(i, j, true);
                    if (has_suitable_rotamers)
                    {
                        pair_hit_probabilities[j].push_back(
                                                            probabilities_swapped[j]);
                        allowed_set1.insert(pair_vector[i].second);
                        allowed_set2.insert(pair_vector[i].first);

                        if (parameters.verbose)
                        {
                            if ((motif.get_residue_id_set().has(id1)
                                    and motif.get_residue_id_set().has(id2))
                                    and (id1
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).first))
                                    and (id2
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).second)))

                                cout << "Adding pair" << endl;
                        }
                    }
                    else
                    {
                        if (parameters.verbose)
                        {
                            if ((motif.get_residue_id_set().has(id1)
                                    and motif.get_residue_id_set().has(id2))
                                    and (id1
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).first))
                                    and (id2
                                            == motif.get_residue_id(
                                                                    motif.get_residue_pair(
                                                                                           j).second)))

                                cout << "Removing pair: no suitable rotamers "
                                        << endl;
                        }
                        removed_count++;
                    }
                }
                else
                {
                    if (parameters.verbose)
                    {
                        if ((motif.get_residue_id_set().has(id1)
                                and motif.get_residue_id_set().has(id2))
                                and (id1
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       j).first))
                                and (id2
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       j).second)))
                            cout << "Probability too low." << endl;
                    }
                }
            }
        }
        HashSet<unsigned int>::iterator hit;
        for (hit = allowed_sites[motif.get_residue_pair(j).first].begin(); hit
                != allowed_sites[motif.get_residue_pair(j).first].end();)
        {
            if (!allowed_set1.has(*hit))
            {
                allowed_sites[motif.get_residue_pair(j).first].erase(hit++);
            }
            else
                hit++;
        }
        for (hit = allowed_sites[motif.get_residue_pair(j).second].begin(); hit
                != allowed_sites[motif.get_residue_pair(j).second].end();)
        {
            if (!allowed_set2.has(*hit))
            {
                allowed_sites[motif.get_residue_pair(j).second].erase(hit++);
            }
            else
                hit++;
        }

    }
    if (parameters.verbose)
        cout << "removed " << removed_count
                << " pairs without suitable rotamers of a total "
                << total_count << endl;
}

bool MotifSearch::determine_pair_hit_rotamers(unsigned int i, unsigned int j,
                                              bool swap)
{
    pair<unsigned int, unsigned int> current_motif_pair =
            motif.get_residue_pair(j), current_hit_pair;
    if (swap)
    {
        current_hit_pair
                = pair<unsigned int, unsigned int> (pair_vector[i].second,
                                                    pair_vector[i].first);
    }
    else
    {
        current_hit_pair = pair_vector[i];
    }

    // get the positions of 2 backbone atoms and CB for each motif hit residue, so we can
    // calculate a transformation matrix that maps the motif rotamers onto the backbone
    // at these positions
    Vector3 CA1, CB1, C1, CA_rot1, CB_rot1, C_rot1, CA2, CB2, C2, CA_rot2,
            CB_rot2, C_rot2;

    CA1 = all_residues[current_hit_pair.first]->getAtom("CA")->getPosition();
    C1 = all_residues[current_hit_pair.first]->getAtom("C")->getPosition();
    if (all_residues[current_hit_pair.first]->getName() != "GLY")
    {
        CB1
                = all_residues[current_hit_pair.first]->getAtom("CB")->getPosition();
    }
    else
    {
        CB1
                = scsel_util::get_ideal_cb_position_for_gly(
                                                all_residues[current_hit_pair.first]->getAtom(
                                                                                              "N")->getPosition(),
                                                C1, CA1);
    }
    // get positions of the backbone atoms of the current rotamers
    CA_rot1
            = motif.get_rotamer(current_motif_pair.first, 0)->getAtom("CA")->getPosition();
    CB_rot1
            = motif.get_rotamer(current_motif_pair.first, 0)->getAtom("CB")->getPosition();
    C_rot1
            = motif.get_rotamer(current_motif_pair.first, 0)->getAtom("C")->getPosition();
    Matrix4x4 T;
    StructureMapper mapper;

    // compute a transformation that maps the current rotamer set onto the backbone at the
    // position specified by the motif hit
    T = mapper.matchPoints(CA_rot1, CB_rot1, C_rot1, CA1, CB1, C1);
    mapper.setTransformation(T);

    // loop over all possible rotamers for the current motif residue
    for (unsigned int k = 0; k
            < motif.get_number_of_rotamers_for_residue(current_motif_pair.first); ++k)
    {
        motif.get_rotamer(current_motif_pair.first, k)->apply(mapper);
    }

    CA2 = all_residues[current_hit_pair.second]->getAtom("CA")->getPosition();
    C2 = all_residues[current_hit_pair.second]->getAtom("C")->getPosition();
    if (all_residues[current_hit_pair.second]->getName() != "GLY")
    {
        CB2
                = all_residues[current_hit_pair.second]->getAtom("CB")->getPosition();
    }
    else
    {
        CB2
                = scsel_util::get_ideal_cb_position_for_gly(
                                                all_residues[current_hit_pair.second]->getAtom(
                                                                                               "N")->getPosition(),
                                                C2, CA2);
    }

    // get positions of the backbone atoms of the current rotamers
    CA_rot2
            = motif.get_rotamer(current_motif_pair.second, 0)->getAtom("CA")->getPosition();
    CB_rot2
            = motif.get_rotamer(current_motif_pair.second, 0)->getAtom("CB")->getPosition();
    C_rot2
            = motif.get_rotamer(current_motif_pair.second, 0)->getAtom("C")->getPosition();

    // compute a transformation that maps the current rotamer set onto the backbone at the
    // position specified by the motif hit
    T = mapper.matchPoints(CA_rot2, CB_rot2, C_rot2, CA2, CB2, C2);
    mapper.setTransformation(T);

    // loop over all possible rotamers for the current motif residue
    for (unsigned int k = 0; k
            < motif.get_number_of_rotamers_for_residue(
                                                       current_motif_pair.second); ++k)
    {
        motif.get_rotamer(current_motif_pair.second, k)->apply(mapper);
    }

    Vector3 catatom11, catatom12, catatom13, catatom21, catatom22, catatom23;

    vector<pair<unsigned int, unsigned int> > current_pair_hit_rotamers;
    vector<double> current_pair_hit_rotamer_scores;
    // loop over all possible rotamers for the current motif residue pair
    for (unsigned int i1 = 0; i1
            < motif.get_number_of_rotamers_for_residue(current_motif_pair.first); ++i1)
    {
        // get the positions of the catalytic atoms of the current rotamer for the first residue
        Residue rot = *motif.get_rotamer(current_motif_pair.first, i1);
        Atom
                a =
                        *rot.getAtom(
                                     motif.get_catalytic_atom_name(
                                                                   current_motif_pair.first,
                                                                   1));
        Vector3 v = a.getPosition();
        catatom11
                = motif.get_rotamer(current_motif_pair.first, i1)->getAtom(
                                                                           motif.get_catalytic_atom_name(
                                                                                                         current_motif_pair.first,
                                                                                                         0))->getPosition();
        catatom12
                = motif.get_rotamer(current_motif_pair.first, i1)->getAtom(
                                                                           motif.get_catalytic_atom_name(
                                                                                                         current_motif_pair.first,
                                                                                                         1))->getPosition();
        catatom13
                = motif.get_rotamer(current_motif_pair.first, i1)->getAtom(
                                                                           motif.get_catalytic_atom_name(
                                                                                                         current_motif_pair.first,
                                                                                                         2))->getPosition();

        for (unsigned int i2 = 0; i2
                < motif.get_number_of_rotamers_for_residue(
                                                           current_motif_pair.second); ++i2)
        {
            // get the positions of the catalytic atoms of the current rotamer for the first residue
            catatom21
                    = motif.get_rotamer(current_motif_pair.second, i2)->getAtom(
                                                                                motif.get_catalytic_atom_name(
                                                                                                              current_motif_pair.second,
                                                                                                              0))->getPosition();
            catatom22
                    = motif.get_rotamer(current_motif_pair.second, i2)->getAtom(
                                                                                motif.get_catalytic_atom_name(
                                                                                                              current_motif_pair.second,
                                                                                                              1))->getPosition();
            catatom23
                    = motif.get_rotamer(current_motif_pair.second, i2)->getAtom(
                                                                                motif.get_catalytic_atom_name(
                                                                                                              current_motif_pair.second,
                                                                                                              2))->getPosition();

            double distance_score = 0;
            double torsion_score = 0;

            double rot_length = (catatom11 - catatom21).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       0, 0)),
                           2);
            rot_length = (catatom11 - catatom22).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       0, 1)),
                           2);
            rot_length = (catatom11 - catatom23).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       0, 2)),
                           2);
            rot_length = (catatom12 - catatom21).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       1, 0)),
                           2);
            rot_length = (catatom12 - catatom22).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       1, 1)),
                           2);
            rot_length = (catatom12 - catatom23).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       1, 2)),
                           2);
            rot_length = (catatom13 - catatom21).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       2, 0)),
                           2);
            rot_length = (catatom13 - catatom22).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       2, 1)),
                           2);
            rot_length = (catatom13 - catatom23).getLength();
            distance_score
                    += pow(
                           (rot_length
                                   - motif.get_catalytic_atom_distance(
                                                                       current_motif_pair.first,
                                                                       current_motif_pair.second,
                                                                       2, 2)),
                           2);

            double torsion = getTorsionAngle(catatom11.x, catatom11.y,
                                             catatom11.z, catatom12.x,
                                             catatom12.y, catatom12.z,
                                             catatom21.x, catatom21.y,
                                             catatom21.z, catatom22.x,
                                             catatom22.y, catatom22.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      0, 1, 0,
                                                                      1)), 2);
            torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z,
                                      catatom12.x, catatom12.y, catatom12.z,
                                      catatom21.x, catatom21.y, catatom21.z,
                                      catatom23.x, catatom23.y, catatom23.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      0, 1, 0,
                                                                      2)), 2);
            torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z,
                                      catatom13.x, catatom13.y, catatom13.z,
                                      catatom21.x, catatom21.y, catatom21.z,
                                      catatom22.x, catatom22.y, catatom22.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      0, 2, 0,
                                                                      1)), 2);
            torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z,
                                      catatom13.x, catatom13.y, catatom13.z,
                                      catatom21.x, catatom21.y, catatom21.z,
                                      catatom23.x, catatom23.y, catatom23.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      0, 2, 0,
                                                                      2)), 2);
            torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z,
                                      catatom12.x, catatom12.y, catatom12.z,
                                      catatom22.x, catatom22.y, catatom22.z,
                                      catatom23.x, catatom23.y, catatom23.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      0, 1, 1,
                                                                      2)), 2);
            torsion = getTorsionAngle(catatom11.x, catatom11.y, catatom11.z,
                                      catatom13.x, catatom13.y, catatom13.z,
                                      catatom22.x, catatom22.y, catatom22.z,
                                      catatom23.x, catatom23.y, catatom23.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      0, 2, 1,
                                                                      2)), 2);
            torsion = getTorsionAngle(catatom12.x, catatom12.y, catatom12.z,
                                      catatom13.x, catatom13.y, catatom13.z,
                                      catatom21.x, catatom21.y, catatom21.z,
                                      catatom22.x, catatom22.y, catatom22.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      1, 2, 0,
                                                                      1)), 2);
            torsion = getTorsionAngle(catatom12.x, catatom12.y, catatom12.z,
                                      catatom13.x, catatom13.y, catatom13.z,
                                      catatom21.x, catatom21.y, catatom21.z,
                                      catatom23.x, catatom23.y, catatom23.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      1, 2, 0,
                                                                      2)), 2);
            torsion = getTorsionAngle(catatom12.x, catatom12.y, catatom12.z,
                                      catatom13.x, catatom13.y, catatom13.z,
                                      catatom22.x, catatom22.y, catatom22.z,
                                      catatom23.x, catatom23.y, catatom23.z);
            torsion_score
                    += pow(
                           (torsion
                                   - motif.get_catalytic_atom_torsion(
                                                                      current_motif_pair.first,
                                                                      current_motif_pair.second,
                                                                      1, 2, 1,
                                                                      2)), 2);

            double pair_score = (torsion_score + distance_score);
            if (sqrt(pair_score / 18.0)
                    < motif.get_maximal_catalytic_geometry_penalty())
            {
                pair<unsigned int, unsigned int> pp(i1, i2);
                current_pair_hit_rotamers.push_back(pp);
                current_pair_hit_rotamer_scores.push_back(pair_score);
            }
        }
    }
    if (current_pair_hit_rotamers.size() > 0)
    {
        // found a hit, remember it, it's suitable rotamers and its partial score probabilities
        pair_hits[j].push_back(current_hit_pair);
        pair_hit_rotamers[j].push_back(current_pair_hit_rotamers);
        pair_hit_rotamer_scores[j].push_back(current_pair_hit_rotamer_scores);
        return true;
    }
    return false;
}

/*
 * recursion, looks for cliques whose sizes equal the number of residue pairs in the motif.
 * these cliques represent a complete motif hit (i.e., a residue in the structure is found
 * for each residue in the moitf, and these residues have a CA/CB geometry similar to the ones in the motif)
 * it tests if the node v can be added to the clique and does that if so
 */
void MotifSearch::clique_search(const Graph& graph, const LabelMap& labelMap,
                                vector<unsigned int>& clique, unsigned int v,
                                vector<vector<unsigned int> >& complete_cliques)
{
    // if v has too few adjacent nodes (i.e. compatible hits), it cannot be in a complete motif,
    // so the clique search will be stopped
    if (out_degree(v, graph) >= motif.get_residue_pair_count() - 1)
    {
        // check if together with v the clique is complete(i.e. each motif residue pair is assigned hit)
        if (clique.size() == motif.get_residue_pair_count() - 1)
        {
            clique.push_back(v);
            // remember the complete clique
            complete_cliques.push_back(clique);
            clique.pop_back();
            return;
        }

        // The set "candidates" should contain all nodes that have connections to all nodes in the clique
        // and represent a pair hit for the next residue pair of the motif (i.e. have a label that is
        // the label of v + 1). These can be part of a larger clique
        set<unsigned int> candidates;
        vector<unsigned int>::iterator cliq_it;
        EdgeIterator edgit = out_edges(v, graph).first;
        for (; edgit != out_edges(v, graph).second; ++edgit)
        {
            if (labelMap[target(*edgit, graph)] == labelMap[v] + 1)
            {
                bool is_candidate = true;
                for (cliq_it = clique.begin(); cliq_it != clique.end(); ++cliq_it)
                {
                    if (!edge(*cliq_it, target(*edgit, graph), graph).second)
                    {
                        is_candidate = false;
                        break;
                    }
                }
                if (is_candidate)
                {
                    candidates.insert(target(*edgit, graph));
                }
            }
        }

        // If there are no candidates, v cannot be part of a clique
        if (candidates.size() == 0)
        {
            return;
        }

        // there are enough candidates, so add v to the putative clique
        // and recurse for each candidate
        else
        {
            // add v to clique
            clique.push_back(v);
            set<unsigned int>::iterator setit = candidates.begin();
            for (; setit != candidates.end(); ++setit)
            {
                // recurse and check if the candidate leads to a complete motif hit clique
                clique_search(graph, labelMap, clique, *setit, complete_cliques);
            }

            // all clique possibilities for cliques containing v are tested, so remove it from the clique
            clique.pop_back();
        }
    }
}

/*
 * searches for complete motif hits (a set of residues in the structure with one residue for each
 * resiude of the motif, that have similar CA/CB geometries as the motif).
 * This is done by creating a compatibility graph for the pair hits and
 * looking for cliques in it, the cliques must contain as many nodes as there are residue
 * pairs in the motif
 */
void MotifSearch::search_complete_motifs(
                                         vector<unsigned int> const &sorted_motif_indices)
{

    motif_hits.clear();

    // if there is only one residue pair in the motif, we do not need to
    // create the compatibility graph, because every pair hit is also a complete motif hit
    if (motif.get_residue_pair_count() == 1)
    {
        unsigned int j;
        for (j = 0; j < pair_hits[0].size(); ++j)
        {
            MotifHit mh(system.getName(), all_residues, motif);
            mh.set_residue_index(0, pair_hits[0][j].first);
            mh.set_residue_index(1, pair_hits[0][j].second);

            set<unsigned int> set1, set2;
            vector<pair<unsigned int, unsigned int> >::iterator it =
                    pair_hit_rotamers[0][j].begin();
            for (; it != pair_hit_rotamers[0][j].end(); ++it)
            {
                set1.insert(it->first);
                set2.insert(it->second);
            }

            mh.insert_allowed_rotamers(0, set1);
            mh.insert_allowed_rotamers(1, set2);

            motif_hits.push_back(mh);
        }
        return;
    }

    vector<pair<unsigned int, unsigned int> > motif_pair_indices;

    // Create the compatibility graph
    Graph graph;
    LabelMap labelMap = get(pair_label_t(), graph);

    unsigned int size1 = 0;
    unsigned int no_overlap_count = 0, kruecke = 0;
    // loop over all motif residue pairs
    for (unsigned int i1 = 0; i1 < motif.get_residue_pair_count(); ++i1)
    {
        unsigned int res11 =
                motif.get_residue_pair(sorted_motif_indices[i1]).first;
        unsigned int res12 =
                motif.get_residue_pair(sorted_motif_indices[i1]).second;

        // loop over all hits for the current motif residue pair
        for (unsigned int j1 = 0; j1
                < pair_hits[sorted_motif_indices[i1]].size(); ++j1)
        {
            // this will hold the index of the motif pair that the current node (i.e. pair hit) is a hit for,
            // and the number of hits there are for the previous motif pairs, so we can later
            // reconstruct which hit in "hits" this node corresponds to
            motif_pair_indices.push_back(
                                         pair<unsigned int, unsigned int> (
                                                                           sorted_motif_indices[i1],
                                                                           size1));

            unsigned int size2 = size1
                    + pair_hits[sorted_motif_indices[i1]].size();

            // loop over all motif residue pairs "after" res11/res12 to avoid looking at each pair combination twice
            for (unsigned int i2 = i1 + 1; i2 < pair_hits.size(); ++i2)
            {
                unsigned int res21 =
                        motif.get_residue_pair(sorted_motif_indices[i2]).first;
                unsigned int res22 =
                        motif.get_residue_pair(sorted_motif_indices[i2]).second;

                // loop over all hits for the second motif residue pair
                for (unsigned int j2 = 0; j2
                        < pair_hits[sorted_motif_indices[i2]].size(); ++j2)
                {
                    bool pairs_of_interest = false;
                    if (parameters.verbose)
                    {
                        String
                                id11 =
                                        (all_residues[pair_hits[sorted_motif_indices[i1]][j1].first])->getID();
                        String
                                id12 =
                                        (all_residues[pair_hits[sorted_motif_indices[i1]][j1].second])->getID();
                        String
                                id21 =
                                        (all_residues[pair_hits[sorted_motif_indices[i2]][j2].first])->getID();
                        String
                                id22 =
                                        (all_residues[pair_hits[sorted_motif_indices[i2]][j2].second])->getID();
                        if ((motif.get_residue_id_set().has(id11)
                                and motif.get_residue_id_set().has(id12)
                                and motif.get_residue_id_set().has(id21)
                                and motif.get_residue_id_set().has(id22))
                                and (id11
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       sorted_motif_indices[i1]).first))
                                and (id12
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       sorted_motif_indices[i1]).second))
                                and (id21
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       sorted_motif_indices[i2]).first))
                                and (id22
                                        == motif.get_residue_id(
                                                                motif.get_residue_pair(
                                                                                       sorted_motif_indices[i2]).second)))
                        {
                            cout << "Potential edge between " << id11
                                    << (all_residues[pair_hits[sorted_motif_indices[i1]][j1].first])->getChain()->getName()
                                    << "/" << id12
                                    << (all_residues[pair_hits[sorted_motif_indices[i1]][j1].second])->getChain()->getName()
                                    << " ("
                                    << motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   sorted_motif_indices[i1]).first)
                                    << "/"
                                    << motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   sorted_motif_indices[i1]).second)
                                    << ") and " << id21
                                    << (all_residues[pair_hits[sorted_motif_indices[i2]][j2].first])->getChain()->getName()
                                    << "/" << id22
                                    << (all_residues[pair_hits[sorted_motif_indices[i2]][j2].second])->getChain()->getName()
                                    << " ("
                                    << motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   sorted_motif_indices[i2]).first)
                                    << "/"
                                    << motif.get_residue_id(
                                                            motif.get_residue_pair(
                                                                                   sorted_motif_indices[i2]).second)
                                    << ")" << endl;
                            pairs_of_interest = true;
                        }
                    }
                    bool compatible = true;

                    // check if the two hits are compatible, i.e. if they have motif residues in common,
                    // these should be assigned the same target attachment sites, and their
                    if (res11 == res21)
                    {
                        if (pair_hits[sorted_motif_indices[i1]][j1].first
                                != pair_hits[sorted_motif_indices[i2]][j2].first)
                        {
                            compatible = false;
                        }
                        else
                        {
                            bool overlap = false;
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it1 =
                                            pair_hit_rotamers[sorted_motif_indices[i1]][j1].begin();
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it2;
                            for (; phr_it1
                                    != pair_hit_rotamers[sorted_motif_indices[i1]][j1].end(); ++phr_it1)
                            {
                                phr_it2
                                        = pair_hit_rotamers[sorted_motif_indices[i2]][j2].begin();
                                for (; phr_it2
                                        != pair_hit_rotamers[sorted_motif_indices[i2]][j2].end(); ++phr_it2)
                                {
                                    if (phr_it1->first == phr_it2->first)
                                    {
                                        overlap = true;
                                        break;
                                    }
                                }
                                if (overlap)
                                    break;
                            }
                            if (!overlap)
                            {
                                if (parameters.verbose && pairs_of_interest)
                                    cout << "No overlapping rotamers!" << endl;
                                no_overlap_count++;
                                compatible = false;
                            }
                        }
                    }
                    else if (res11 == res22)
                    {
                        if (pair_hits[sorted_motif_indices[i1]][j1].first
                                != pair_hits[sorted_motif_indices[i2]][j2].second)
                        {
                            compatible = false;
                        }
                        else
                        {
                            bool overlap = false;
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it1 =
                                            pair_hit_rotamers[sorted_motif_indices[i1]][j1].begin();
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it2;
                            for (; phr_it1
                                    != pair_hit_rotamers[sorted_motif_indices[i1]][j1].end(); ++phr_it1)
                            {
                                phr_it2
                                        = pair_hit_rotamers[sorted_motif_indices[i2]][j2].begin();
                                for (; phr_it2
                                        != pair_hit_rotamers[sorted_motif_indices[i2]][j2].end(); ++phr_it2)
                                {
                                    if (phr_it1->first == phr_it2->second)
                                    {
                                        overlap = true;
                                        break;
                                    }
                                }
                                if (overlap)
                                    break;
                            }
                            if (!overlap)
                            {
                                if (parameters.verbose && pairs_of_interest)
                                    cout << "No overlapping rotamers!" << endl;
                                no_overlap_count++;
                                compatible = false;
                            }
                        }
                    }
                    else if (res12 == res21)
                    {
                        if (pair_hits[sorted_motif_indices[i1]][j1].second
                                != pair_hits[sorted_motif_indices[i2]][j2].first)
                        {
                            compatible = false;
                        }
                        else
                        {
                            bool overlap = false;
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it1 =
                                            pair_hit_rotamers[sorted_motif_indices[i1]][j1].begin();
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it2;
                            for (; phr_it1
                                    != pair_hit_rotamers[sorted_motif_indices[i1]][j1].end(); ++phr_it1)
                            {
                                phr_it2
                                        = pair_hit_rotamers[sorted_motif_indices[i2]][j2].begin();
                                for (; phr_it2
                                        != pair_hit_rotamers[sorted_motif_indices[i2]][j2].end(); ++phr_it2)
                                {
                                    if (phr_it1->second == phr_it2->first)
                                    {
                                        overlap = true;
                                        break;
                                    }
                                }
                                if (overlap)
                                    break;
                            }
                            if (!overlap)
                            {
                                if (parameters.verbose && pairs_of_interest)
                                    cout << "No overlapping rotamers!";
                                no_overlap_count++;
                                compatible = false;
                            }
                        }
                    }
                    else if (res12 == res22)
                    {
                        if (pair_hits[sorted_motif_indices[i1]][j1].second
                                != pair_hits[sorted_motif_indices[i2]][j2].second)
                        {
                            compatible = false;
                        }
                        else
                        {
                            bool overlap = false;
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it1 =
                                            pair_hit_rotamers[sorted_motif_indices[i1]][j1].begin();
                            vector<pair<unsigned int, unsigned int> >::iterator
                                    phr_it2;
                            for (; phr_it1
                                    != pair_hit_rotamers[sorted_motif_indices[i1]][j1].end(); ++phr_it1)
                            {
                                phr_it2
                                        = pair_hit_rotamers[sorted_motif_indices[i2]][j2].begin();
                                for (; phr_it2
                                        != pair_hit_rotamers[sorted_motif_indices[i2]][j2].end(); ++phr_it2)
                                {
                                    if (phr_it1->second == phr_it2->second)
                                    {
                                        overlap = true;
                                        break;
                                    }
                                }
                                if (overlap)
                                    break;
                            }
                            if (!overlap)
                            {
                                if (parameters.verbose && pairs_of_interest)
                                    cout << "No overlapping rotamers!";
                                no_overlap_count++;
                                compatible = false;
                            }
                        }
                    }
                    else
                    {
                        Vector3
                                v11 =
                                        (all_residues[pair_hits[sorted_motif_indices[i1]][j1].first])->getAtom(
                                                                                                               "CA")->getPosition();
                        Vector3
                                v12 =
                                        (all_residues[pair_hits[sorted_motif_indices[i1]][j1].second])->getAtom(
                                                                                                                "CA")->getPosition();
                        Vector3
                                v21 =
                                        (all_residues[pair_hits[sorted_motif_indices[i2]][j2].first])->getAtom(
                                                                                                               "CA")->getPosition();
                        Vector3
                                v22 =
                                        (all_residues[pair_hits[sorted_motif_indices[i2]][j2].second])->getAtom(
                                                                                                                "CA")->getPosition();
                        if ((v11 - v21).getLength()
                                > motif.get_maximal_CA_distance()
                                or (v11 - v21).getLength()
                                        < motif.get_minimal_CA_distance()
                                or (v11 - v22).getLength()
                                        > motif.get_maximal_CA_distance()
                                or (v11 - v22).getLength()
                                        < motif.get_minimal_CA_distance()
                                or (v12 - v21).getLength()
                                        > motif.get_maximal_CA_distance()
                                or (v12 - v21).getLength()
                                        < motif.get_minimal_CA_distance()
                                or (v12 - v22).getLength()
                                        > motif.get_maximal_CA_distance()
                                or (v12 - v22).getLength()
                                        < motif.get_minimal_CA_distance())
                        {
                            compatible = false;
                            kruecke++;
                        }
                    }
                    // if the two hits are compatible, add an edge between the two respective nodes to the graph
                    if (compatible)
                    {
                        if (parameters.verbose && pairs_of_interest)
                            cout << "Edge added." << endl;
                        add_edge(size1 + j1, size2 + j2, graph);
                        //label the nodes, so we know which residue pair the node represents a pair hit for
                        put(labelMap, size1 + j1, i1);
                        put(labelMap, size2 + j2, i2);
                    }
                    else if (parameters.verbose && pairs_of_interest)
                        cout << "Incompatible. No edge added." << endl;
                }
                size2 += pair_hits[sorted_motif_indices[i2]].size();
            }
        }
        // the nodes in the graph have indices according to the order of pair hits in vector "hits".
        // for this purpose, the hits are counted continuously, regardless which motif residue pair
        // they are a hit for. This is accomplished by remembering the amount of hits for the previous motif
        // residue pairs in the size1 / size2 variables
        size1 += pair_hits[sorted_motif_indices[i1]].size();
    }

    VertexIt first = vertices(graph).first;
    VertexIt last = vertices(graph).second;

    EdgeIterator edit;

    bool neighbor_labels[motif.get_residue_pair_count()];

    unsigned int clear_count = 1;
    unsigned int global_clear_count = 0;
    while (clear_count > 0)
    {
        first = vertices(graph).first;
        // start a loop over the nodes of the compatibility graph
        clear_count = 0;
        unsigned int count = 0;
        for (; first != last; ++first)
        {
            if (out_degree(*first, graph) == 0)
                continue;

            // remove vertices with an edge count smaller than motif residue pair count -1,
            // they cannot be part of a complete motif hit clique
            else if (out_degree(*first, graph) < motif.get_residue_pair_count()
                    - 1)
            {
                clear_count++;
                clear_vertex(*first, graph);
                continue;
            }
            for (unsigned int t = 0; t < motif.get_residue_pair_count(); ++t)
            {
                neighbor_labels[t] = false;
            }

            // remove vertices that do not have at least one edge to a vertex of each other "color",
            // i.e. an edge to a pair hit for each other motif residue pair than itself
            neighbor_labels[labelMap[*first]] = true;
            edit = out_edges(*first, graph).first;
            for (; edit != out_edges(*first, graph).second; ++edit)
            {
                if (!neighbor_labels[labelMap[target(*edit, graph)]])
                {
                    neighbor_labels[labelMap[target(*edit, graph)]] = true;
                    count++;
                }
                if (count == motif.get_residue_pair_count() - 1)
                    break;

            }
            if (count < motif.get_residue_pair_count() - 1)
            {
                clear_vertex(*first, graph);
                clear_count++;
            }
            count = 0;
        }
        global_clear_count += clear_count;
        if (parameters.verbose)
            cout << "deleted " << clear_count << " vertices of "
                    << num_vertices(graph) << endl;
    }
    if (parameters.verbose)
    {
        cout << "Compatibility graph created: " << endl;
        cout << "   Prevented " << no_overlap_count
                << " edges with no allowed rotamer set overlap." << endl;
        cout << "   Prevented " << kruecke
                << " edges with CA distance way out of line." << endl;
        cout << "   Prevented " << global_clear_count
                << " vertices with not enough edges." << endl;
        cout << "   Vertices: " << num_vertices(graph) - global_clear_count
                << ", edges: " << num_edges(graph) << endl;
    }

    first = vertices(graph).first;
    last = vertices(graph).second;

    vector<vector<unsigned int> > complete_cliques;

    // this will hold the clique, i.e. part of the potential motif hit
    vector<unsigned int> clique;

    // start a loop over the nodes of the compatibility graph
    for (; first != last; ++first)
    {
        // each clique has to contain exactly one node that represents a hit for the first
        // residue pair in motif_pairs, so we have to start the recursion only with these nodes

        if (labelMap[*first] == 0)
        {
            // if the node has too few edges, it cannot be part of a clique of motif_pairs size
            if (out_degree(*first, graph) >= motif.get_residue_pair_count() - 1)
            {
                // the clique contains of one node at this point, it represents a hit for the first
                // motif residue pair
                clique.push_back(*first);

                // candidates for clique members are all nodes connected by an edge to the
                // one node in the clique
                set<unsigned int> candidates;
                EdgeIterator it = out_edges(*first, graph).first;
                for (; it != out_edges(*first, graph).second; ++it)
                {
                    if (labelMap[target(*it, graph)] == labelMap[*first] + 1)
                    {
                        //						unsigned int i1 = target(*it, graph);
                        candidates.insert(target(*it, graph));
                    }
                }

                // for each node that is a candidate we recurse and check if it is in one ore more complete cliques,
                // i.e. in one ore more complete motif hits.
                // after that, we have dealt with this candidate and can remove it from the candidate list,
                // so avoiding to find a motif hit several times
                for (set<unsigned int>::iterator setit = candidates.begin(); setit
                        != candidates.end();)
                {
                    clique_search(graph, labelMap, clique, *setit,
                                  complete_cliques);
                    // all possibilites involving this candidate ahve been tested, so we can remove it from the candidates set
                    candidates.erase(setit++);
                }
                // we checked all clique possibilities for this node, so remove it
                clique.pop_back();
            }
        }
        else
            break;
    }
    if (parameters.verbose)
        cout << "Found " << complete_cliques.size() << " clique(s). " << endl;

    // loop over all complete cliques and create the corresponding MotifHit instances
    for (vector<vector<unsigned int> >::iterator it = complete_cliques.begin(); it
            != complete_cliques.end(); ++it)
    {
        // this vector will hold the indices of the residues in the structure that form the motif hit
        // each index is in the same vector position as the motif residue it represents is in the motif vector
        MotifHit motif_hit(system.getName(), all_residues, motif);
        bool is_wt_hit = true;

        // loop over the hits/nodes in the current clique
        for (vector<unsigned int>::iterator itt = it->begin(); itt != it->end(); ++itt)
        {
            // the current node is a hit for the motif residue pair i
            unsigned int i = motif_pair_indices[*itt].first;

            // retrieve the pair of indices that point to the residues of the hit
            pair<unsigned int, unsigned int> p =
                    pair_hits[motif_pair_indices[*itt].first][*itt
                            - motif_pair_indices[*itt].second];
            vector<pair<unsigned int, unsigned int> >::iterator pr_it =
                    pair_hit_rotamers[motif_pair_indices[*itt].first][*itt
                            - motif_pair_indices[*itt].second].begin();
            set<unsigned int> s1, s2;
            for (unsigned int j = 0; j
                    < pair_hit_rotamers[motif_pair_indices[*itt].first][*itt
                            - motif_pair_indices[*itt].second].size(); ++j)
            {
                motif_hit.set_pair_rotamer_score(
                                                 i,
                                                 pair_hit_rotamers[motif_pair_indices[*itt].first][*itt
                                                         - motif_pair_indices[*itt].second][j].first,
                                                 pair_hit_rotamers[motif_pair_indices[*itt].first][*itt
                                                         - motif_pair_indices[*itt].second][j].second,
                                                 pair_hit_rotamer_scores[motif_pair_indices[*itt].first][*itt
                                                         - motif_pair_indices[*itt].second][j]);
                s1.insert(
                          pair_hit_rotamers[motif_pair_indices[*itt].first][*itt
                                  - motif_pair_indices[*itt].second][j].first);
                s2.insert(
                          pair_hit_rotamers[motif_pair_indices[*itt].first][*itt
                                  - motif_pair_indices[*itt].second][j].second);
            }

            if (parameters.verbose && is_wt_hit)
            {
                String id1 = all_residues[p.first]->getID();
                String id2 = all_residues[p.second]->getID();
                if (!((motif.get_residue_id_set().has(id1)
                        and motif.get_residue_id_set().has(id2))
                        and (id1
                                == motif.get_residue_id(
                                                        motif.get_residue_pair(
                                                                               i).first))
                        and (id2
                                == motif.get_residue_id(
                                                        motif.get_residue_pair(
                                                                               i).second))))
                {
                    is_wt_hit = false;
                }
            }

            motif_hit.insert_allowed_rotamers(motif.get_residue_pair(i).first,
                                              s1);
            motif_hit.insert_allowed_rotamers(motif.get_residue_pair(i).second,
                                              s2);

            // remember the indices of the residues that form the complete motif hit at the correct positions
            motif_hit.set_residue_index(motif.get_residue_pair(i).first,
                                        p.first);
            motif_hit.set_residue_index(motif.get_residue_pair(i).second,
                                        p.second);
        }
        motif_hit.is_wt_hit = false;
        if (parameters.verbose && is_wt_hit)
        {
            motif_hit.is_wt_hit = true;
            cout << "hit : " << motif_hit.get_ID() << endl;
        }
        motif_hits.push_back(motif_hit);
    }
    pair_vector.clear();
    pair_hit_probabilities.clear();
    pair_hit_rotamer_scores.clear();
    pair_hits.clear();
    pair_hit_rotamers.clear();
}

/*
 * checks if there is an atom that fits the BALL expression of an interaction center
 * closer than the cutoff radius around the reconstructed center position. the center position
 * is recontructed using the reconstruction matrix
 */
void MotifSearch::check_interaction_centers()
{
    // create a Hashgrid that will hold all potential interaction center atoms
    HashGrid3<Atom*>
            interaction_atoms_grid(
                                   lower
                                           - Vector3(
                                                     motif.get_maximal_interaction_cutoff_radius()),
                                   upper - lower
                                           + Vector3(
                                                     2.0
                                                             * motif.get_maximal_interaction_cutoff_radius()),
                                   motif.get_maximal_interaction_cutoff_radius());

    AtomIterator ait;
    unsigned int i;

    // loop over all atoms of the protein
    for (vector<Residue*>::iterator res_it = all_residues.begin(); res_it
            != all_residues.end(); ++res_it)
    {
        ait = (*res_it)->beginAtom();
        for (; ait != (*res_it)->endAtom(); ++ait)
        {
            // loop over all interaction centers
            for (i = 0; i < motif.get_interaction_center_count(); ++i)
            {
                // if the atom fits one of the interaction center BALL expressions, insert itinto the grid
                if (motif.get_interaction_center_expression(i)(*ait))
                {
                    interaction_atoms_grid.insert(ait->getPosition(), &*ait);
                }
            }
        }
    }

    // loop over all motif hits
    list<MotifHit>::iterator mit = motif_hits.begin();

    while (mit != motif_hits.end())
    {
        bool all_centers =
                mit->check_interaction_centers(interaction_atoms_grid, false);
        if (!all_centers)
        {
            mit = motif_hits.erase(mit);
        }
        else
        {
            mit++;
        }
    }
}

/*
 * checks complete motif hits, that map a residue of the structure to a residue in the motif, if the SAS
 * of the structure residues are similar to the SAS wanted for the motif residues,
 * and add a penalty according to the discrepancy to the probablity score of the motif hit
 */
void MotifSearch::compute_relative_SAS_score()
{

    // Assign vdW radii to the Atoms in the protein
    AssignRadiusProcessor rp(parameters.get_ball_directory()
            + "/data/radii/PARSE.siz");
    system.apply(rp);

    // calculate the SAS area for each atom in the protein and store it in a Hashmap
    HashMap<const Atom*, float> atom_areas;
    calculateSASAtomAreas(system, atom_areas, 1.5, 400);

    // will contain the relative solvent accessible surface area of each residue in all_residue at
    // the same index
    vector<double> residue_SAS_areas;

    AtomIterator atom_it;
    // compute the relative SAS areas of all residues
    for (vector<Residue*>::iterator res_it = all_residues.begin(); res_it
            != all_residues.end(); ++res_it)
    {
        double residue_area = 0.0;

        // compute the SASA of the residue by adding the SASAs of the Atoms in it
        for (atom_it = (*res_it)->beginAtom(); +atom_it; ++atom_it)
        {
            residue_area += atom_areas[&*atom_it];
        }
        double reference_SASA = -1.0;
        try
        {
            reference_SASA
                    = parameters.get_reference_SASA((*res_it)->getName());
        }
        catch (Exception::GeneralException e)
        {
        }
        // normalize the SASA by the SASA value of a reference residue of the same amino acid type
        residue_area /= reference_SASA;
        residue_SAS_areas.push_back(residue_area);
    }

    // loop over all complete motif hits
    list<MotifHit>::iterator mh_it = motif_hits.begin();

    while (mh_it != motif_hits.end())
    {
        bool got_SAS_score =
                mh_it->compute_relative_SAS_score(residue_SAS_areas);
        if (!got_SAS_score)
        {
            mh_it = motif_hits.erase(mh_it);
        }
        else
        {
            mh_it++;
        }
    }

}

/*
 * recontructs the positions of all spheres for each motif hit,
 * and checks of there are any backbone atoms inside the spheres.
 * if so, computes a penalty.
 */
void MotifSearch::check_ligand_clashes()
{
    // loop over all motif hits
    list<MotifHit>::iterator mh_it = motif_hits.begin();
    for (; mh_it != motif_hits.end(); ++mh_it)
    {
        mh_it->check_ligand_clashes(backbone_atoms_grid);
    }
}

/*
 * searches for possible motif geometries the PDB System
 * computes the scores for these hits and adds them to the ranking datastructure
 * returns the number of found motif hits
 */
unsigned int MotifSearch::search_for_motifs(Ranking& ranking)
{

    Timer t;
    if (parameters.time_is_of_the_essence)
        t.start();

    HashSet<Residue*> secondary_structure_residues =
            get_secondary_structure_residues();

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime()
                << " seconds for secondary structure determination" << endl;
        t.reset();
        t.start();
    }

    HashSet<Residue*> pocket_residues;
    if (parameters.get_compute_surface_pockets())
    {
        // determine which residues are forming a pocket on the protein surface ore are in the
        // vicinity of a pocket and therefore eligible to be attachment sites
        PocketFinder pocket_finder(system, parameters);
        pocket_residues = pocket_finder.get_pocket_residues();
    }

    if (parameters.verbose)
    {
        cout << "Pocket residues:";
        HashSet<Residue*>::iterator it = pocket_residues.begin();
        for (; it != pocket_residues.end(); ++it)
        {
            cout << "" << (*it)->getID() << ":" << (*it)->getChain()->getName()
                    << ", ";
        }
        cout << endl;
    }
    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for pocket determination" << endl;
        t.reset();
        t.start();
    }

    // compute all Residue pairs that could potentially be motif pairs
    // because their Ca-Range is in the allowed interval
    get_residue_pairs(pocket_residues, secondary_structure_residues);

    if (parameters.verbose)
        cout << "Considering " << all_residues.size() << " of "
                << system.countResidues() << " Residues" << endl;

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for Ca pair calculation" << endl;
        t.reset();
        t.start();
    }

    // computes all residue pairs in the structure that are potential hits
    // for a residue pair of the motif
    compute_pair_hits();

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for pair hit calculation" << endl;
        t.reset();
    }

    // create a vector sorted_pairs, in which the residue pairs of the motif are sorted by the
    // number of hits that were found for them, the index of the pair in member variable pair_vector.
    // with the least hits is in sorted_pairs[0] etc...
    vector<unsigned int> sorted_indices;
    bool used_indices[motif.get_residue_pair_count()];
    unsigned int i;
    for (i = 0; i < motif.get_residue_pair_count(); ++i)
    {
        used_indices[i] = false;
    }

    for (unsigned int j = 0; j < pair_hits.size(); ++j)
    {
        unsigned int min = Limits<int>::max();
        int minindex = -1;
        for (i = 0; i < pair_hits.size(); ++i)
        {
            if (used_indices[i])
                continue;
            if (pair_hits[i].size() < min)
            {
                min = pair_hits[i].size();
                minindex = i;
            }
        }
        used_indices[minindex] = true;
        sorted_indices.push_back(minindex);
    }

    // if there are no hits for a motif residue pair, then there can be no
    // complete motif hit, and we can skip this structure
    if (pair_hits[sorted_indices[0]].size() == 0)
    {
        cout << "not enough pair hits" << endl;
        return 0;
    }

    if (parameters.time_is_of_the_essence)
        t.start();

    // search for complete motif hits, i.e. if we can assign a residue of the pdb structure
    // to each residue of the motif, so that the residues in the structure
    // have a similar Ca-Cb geometry as the ones in the motif.
    // this is done by building a compatibility graph with the pair hits as nodes,
    // and looking for cliques in that graph
    search_complete_motifs(sorted_indices);

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for compatibility clique search"
                << endl;
        t.reset();
        t.start();
    }

    if (motif_hits.size() == 0)
    {
        return 0;
    }

    // Hashgrid with all backbone atoms for clash checking
    // spacing is the sum of the 2 largest vdW-radii (Sulfur)
    // 					     	   			origin                      size 			 	spacing
    backbone_atoms_grid = HashGrid3<Atom*> (lower - Vector3(0.1F), upper
            - lower + Vector3(0.2F), 3.7F);

    // this is used to determine which atoms should be checked for clashes
    Expression backbone_expression("backbone() OR name(CB)");

    AtomIterator ait;
    // loop over all atoms of the protein
    for (vector<Residue*>::iterator res_it = all_residues.begin(); res_it
            != all_residues.end(); ++res_it)
    {
        ait = (*res_it)->beginAtom();
        for (; ait != (*res_it)->endAtom(); ++ait)
        {
            // if the atom is a backbone atom or a CB atom, insert it into the grid
            if (backbone_expression(*ait))
            {
                backbone_atoms_grid.insert(ait->getPosition(), &*ait);
            }
        }
    }

    // for each motif, construct the motif sidechains onto the anchor positions
    // indicated by the hit, and determine which rotamer combination of the constructed sidechains
    // is most similar to the geometry of the catalytic atoms in the motif
    list<MotifHit>::iterator mit = motif_hits.begin();
    unsigned int clash_rotamer_discarded_hits = 0;
    while (mit != motif_hits.end())
    {
        bool rotamers_determined =
                        mit->construct_motif_rotamers(
                                                      backbone_atoms_grid,
                                                      motif.get_maximal_backbone_clash_penalty());
        if (parameters.verbose && mit->is_wt_hit)
            cout << "Hit " << mit->get_ID();
        if (!rotamers_determined)
        {
            clash_rotamer_discarded_hits++;
            if (parameters.verbose && mit->is_wt_hit)
                cout << " removed: not enough non-clashing rotamers" << endl;
            mit = motif_hits.erase(mit);
        }
        else if (mit->get_catalytic_atoms_geometry_penalty()
                > motif.get_maximal_catalytic_geometry_penalty())
        {
            if (parameters.verbose && mit->is_wt_hit)
                cout << " removed because of bad geom. penalty score: "
                        << mit->get_catalytic_atoms_geometry_penalty() << endl;
            mit = motif_hits.erase(mit);
        }
        else
        {
            if (parameters.verbose && mit->is_wt_hit)
                cout << " geom. penalty score: "
                        << mit->get_catalytic_atoms_geometry_penalty() << endl;
            mit++;
        }
    }

    if (parameters.verbose)
        cout << clash_rotamer_discarded_hits
                << " hits discarded because of no non-clashing rotamers"
                << endl;
    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for rotamer construction" << endl;
        t.reset();
        t.start();
    }

    // if the user wants to compute the scores for sphere clashes or/and
    // interaction centers, compute the recnostruction matrices that map the
    // positions of the catalytic atoms in the motif onto the positions of these atoms in the
    // constructed motif hit. these matrices are needed to compute the positions
    // of the spheres and interaction centers in the motif hit
    if (parameters.get_calculate_ligand_clashes()
            || parameters.get_calculate_interaction_centers())
    {
        mit = motif_hits.begin();
        for (; mit != motif_hits.end(); mit++)
        {
            mit->compute_mapping_matrix();
        }
    }

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for reconstruction matrices"
                << endl;
        t.reset();
        t.start();
    }

    // if the user wants to check for interaction centers, do it
    if (parameters.get_calculate_interaction_centers())
    {
        // interaction centers are atoms of a type that is described in a
        // BALL expression. such an atom has to be present in the PDB structure
        // near the desired position of that atom, or the motif hit will be erased
        check_interaction_centers();
    }

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for interaction centers" << endl;
        t.reset();
        t.start();
    }

    // if the user wants to check for backbone clashes, do that
    // an compute the according penalties
    if (parameters.get_calculate_backbone_clashes())
    {
        mit = motif_hits.begin();
        while (mit != motif_hits.end())
        {
            mit->check_backbone_clashes(backbone_atoms_grid);
            if (mit->get_backbone_clash_penalty()
                    > motif.get_maximal_backbone_clash_penalty())
            {
                if (parameters.verbose && mit->is_wt_hit)
                    cout << "Hit " << mit->get_ID()
                            << " removed because of backbone clashes: "
                            << mit->get_backbone_clash_penalty() << endl;

                mit = motif_hits.erase(mit);
            }
            else
            {
                mit++;
            }
        }
    }

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for backbone clashes" << endl;
        t.reset();
        t.start();
    }

    // if the user wants to check for sphere clashes,
    // do so and compute the penalties
    if (parameters.get_calculate_ligand_clashes())
    {
        check_ligand_clashes();
    }

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for substrate clashes" << endl;
        t.reset();
        t.start();
    }

    // if wanted, compute the relative SAS areas of the anchor residues of a motif hit,
    // and calculate a penalty according to the differences of these to the
    // ones specified in the motif
    if (parameters.get_calculate_relative_SAS())
    {
        compute_relative_SAS_score();
    }

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for SAS scoring" << endl;
        t.reset();
        t.start();
    }

    // add the found motif hits to the ranking lists
    ranking.add_motif_hits(motif_hits);

    if (parameters.time_is_of_the_essence)
    {
        t.stop();
        cout << t.getUserTime() << " seconds for ranking" << endl;
    }

    // return the number of encountered motif hits
    return motif_hits.size();
}
