/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "parameters.h"
using namespace BALL;
using namespace std;


Parameters::Parameters(String const &parameter_file,  bool verb = false,
                                               bool time_is_money = false)
{
	verbose = verb;
	time_is_of_the_essence = time_is_money;
	read_parameters(parameter_file);
	if (calculate_relative_SAS)
	    read_reference_SAS_areas();
}

Parameters::Parameters()
{
}

Parameters::~Parameters()
{
}


/*
 * reads in the parameters for the scaffold search
 */
void Parameters::read_parameters(String const &parameter_file)
{
	INIFile f(parameter_file);
	f.read();
	String value;
	try
	{
		value = f.getValue("scores", "calculate_backbone_clashes");
		if (value == INIFile::UNDEFINED)
		    calculate_backbone_clashes = false;
		else
		    calculate_backbone_clashes = value.toBool();

        value = f.getValue("scores", "calculate_relative_SAS");
        if (value == INIFile::UNDEFINED)
            calculate_relative_SAS = false;
        else
            calculate_relative_SAS = value.toBool();

        value = f.getValue("scores", "calculate_substrate_clashes");
        if (value == INIFile::UNDEFINED)
            calculate_ligand_clashes = false;
        else
            calculate_ligand_clashes = value.toBool();

        value = f.getValue("scores", "calculate_rotamer_probabilities");
        if (value == INIFile::UNDEFINED)
            calculate_rotamer_probabilities = false;
        else
            calculate_rotamer_probabilities = value.toBool();

        value = f.getValue("scores", "calculate_interaction_centers");
        if (value == INIFile::UNDEFINED)
            calculate_interaction_centers = false;
        else
            calculate_interaction_centers = value.toBool();

        structure_directory = f.getValue("paths", "structure_directory");
        if (structure_directory == INIFile::UNDEFINED)
            throw String("structure_directory");

        ball_directory = f.getValue("paths", "ball_directory");
        if (ball_directory == INIFile::UNDEFINED)
            throw String("ball_directory");

        include_list_file = f.getValue("paths", "include_list_file");
        if (include_list_file == INIFile::UNDEFINED)
            throw String("include_list_file");

        reference_SASA_file = f.getValue("paths", "reference_SASA_file");
        if (reference_SASA_file == INIFile::UNDEFINED)
        {
            if (calculate_relative_SAS)
                throw String("reference_SASA_file");
        }

        value = f.getValue("thresholds","minimal_residue_number");
        minimal_residue_number = value.toUnsignedInt();

        value = f.getValue("thresholds","maximal_residue_number");
        maximal_residue_number = value.toUnsignedInt();

        value = f.getValue("thresholds","minimal_resolution");
        minimal_resolution = value.toDouble();

        value = f.getValue("thresholds","minimal_chain_length");
        minimal_chain_length = value.toUnsignedInt();

        value = f.getValue("thresholds","helix_boundary_length");
        helix_boundary_length = value.toUnsignedInt();

        value = f.getValue("thresholds","strand_boundary_length");
        strand_boundary_length = value.toUnsignedInt();

        value = f.getValue("thresholds","ranking_list_length");
        ranking_list_length = value.toUnsignedInt();

        value = f.getValue("ligsitecs", "compute_surface_pockets");
        if (value == INIFile::UNDEFINED)
            compute_surface_pockets = false;
        else
            compute_surface_pockets = value.toBool();

        if (compute_surface_pockets)
        {
            value = f.getValue("ligsitecs","pocket_grid_space");
            pocket_grid_space = value.toFloat();

            value = f.getValue("ligsitecs","number_of_pockets");
            number_of_pockets = value.toUnsignedInt();

            value = f.getValue("ligsitecs","SSSthreshold");
            SSSthreshold = value.toUnsignedInt();

            value = f.getValue("ligsitecs","surface_density");
            surface_density = value.toFloat();

            value = f.getValue("ligsitecs","pocket_radius");
            pocket_radius = value.toFloat();
        }

    }
	catch (String s)
	{
	    String message = "Parameter " + s + " is missing.";
	    Exception::GeneralException e(parameter_file.c_str(), -1,
	                                  "Parameter file", message);
	    throw e;
	}
	catch (int i)
    {
		String message;
    	if (i == 2)
    	{
			message = "PDB structure directory is empty or does not exist";
    	}
    	if (i == 3)
    	{
			message = "Value has to be either \"true\" or \"false\".";
    	}
    	if (i == 4)
    	{
			message = "Value has to be an integer number.";
    	}
    	if (i == 5)
    	{
			message = "Value has to be a floating point number.";
    	}
    	if (i == 6)
	    {
			message = "One or more parameter(s) are missing.";
	    }
    	if (i == 7)
	    {
			message = "Include list file could not be found.";
	    }
	    if (i == 8)
	    {
	    	message = "No files in the structure directory that are "
	    	          "listed in the include list.";
	    }
	    Exception::GeneralException e(parameter_file.c_str(),
	                                  -1, "Parameter file", message);
		throw e;
    }
}

/*
 * reads in the reference SAS areas
 */
void Parameters::read_reference_SAS_areas ()
{
    // this set is used to check if the amino acid names in the file are valid
    set<String> amino_acid_types;
    amino_acid_types.insert("HIS");
    amino_acid_types.insert("ASP");
    amino_acid_types.insert("SER");
    amino_acid_types.insert("GLU");
    amino_acid_types.insert("GLN");
    amino_acid_types.insert("ASN");
    amino_acid_types.insert("TRP");
    amino_acid_types.insert("TYR");
    amino_acid_types.insert("LYS");
    amino_acid_types.insert("ARG");
    amino_acid_types.insert("THR");
    amino_acid_types.insert("ILE");
    amino_acid_types.insert("LEU");
    amino_acid_types.insert("CYS");
    amino_acid_types.insert("MET");
    amino_acid_types.insert("ARG");
    amino_acid_types.insert("PRO");
    amino_acid_types.insert("VAL");
    amino_acid_types.insert("PHE");
    amino_acid_types.insert("ALA");
    amino_acid_types.insert("GLY");

    reference_SAS_areas.clear();
    LineBasedFile f;
    f.open(reference_SASA_file);
    String line;
    unsigned int line_number = 0;
    while(f.readLine())
    {
    	line = f.getLine();
    	line_number++;

    	// if the line is empty or a comment (starting with #), skip it
    	if (line.isWhitespace() || line.size() == 0 || line.getSubstring(0,1) == "#")
    	{
    		continue;
    	}
		if (amino_acid_types.find(line.getField(0)) == amino_acid_types.end())
		{
			String message (line.getField(0) + " is no amino acid.");
			Exception::GeneralException e(reference_SASA_file.c_str(), line_number, "Reference SASA file", message);
			throw e;
		}
		if (!line.getField(1).isFloat())
		{
			String message ("Value has to be a floating point number.");
			Exception::GeneralException e(reference_SASA_file.c_str(), line_number, "Reference SASA file", message);
			throw e;
		}
		reference_SAS_areas[line.getField(0)] = line.getField(1).toDouble();
    }
}

bool Parameters::get_calculate_backbone_clashes() const
{
	return calculate_backbone_clashes;
}

bool Parameters::get_calculate_interaction_centers() const
{
	return calculate_interaction_centers;
}

bool Parameters::get_calculate_relative_SAS() const
{
	return calculate_relative_SAS;
}

bool Parameters::get_calculate_ligand_clashes() const
{
	return calculate_ligand_clashes;
}

bool Parameters::get_calculate_rotamer_probabilities() const
{
	return calculate_rotamer_probabilities;
}

const String Parameters::get_structure_directory() const
{
	return structure_directory;
}

const unsigned int Parameters::get_minimal_residue_number() const
{
	return minimal_residue_number;
}

const unsigned int Parameters::get_maximal_residue_number() const
{
	return maximal_residue_number;
}

const float Parameters::get_minimal_resolution() const
{
	return minimal_resolution;
}

const unsigned int Parameters::get_minimal_chain_length() const
{
	return minimal_chain_length;
}

const unsigned int Parameters::get_strand_boundary_length() const
{
	return strand_boundary_length;
}

const unsigned int Parameters::get_helix_boundary_length() const
{
	return helix_boundary_length;
}

const unsigned int Parameters::get_ranking_list_length() const
{
	return ranking_list_length;
}

const String Parameters::get_include_list_file() const
{
	return include_list_file;
}

double Parameters::get_reference_SASA(const String& amino_acid)
{
	if (reference_SAS_areas.has(amino_acid))
	{
		return reference_SAS_areas[amino_acid];
	}
	else
	{
		Exception::GeneralException e(__FILE__, __LINE__);
		e.setMessage("Error when trying to obtain the reference SASA:"
		             " no amino_acid three letter code.");
		throw e;
	}
}


const String Parameters::get_ball_directory() const
{
	return ball_directory;
}

/**
 * returns spacing of the LIGSITEcs grid
 */
const float Parameters::get_pocket_grid_space() const
{
	return pocket_grid_space;
}


bool Parameters::get_compute_surface_pockets()
{
	return compute_surface_pockets;
}

/**
 * returns how many of the largest pockets found by LIGSITEcs are used by ScaffoldSelection
 */
const unsigned int Parameters::get_number_of_pockets() const
{
	return number_of_pockets;
}

/**
 * returns the amount of SSS events a LIGSITEcs grid point has to have in order to be considered as lying
 * in a pocket
 */
const unsigned int Parameters::get_SSSthreshold() const
{
	return SSSthreshold;
}

/**
 * returns the density of the LIGSITEcs Surface
 */
const float Parameters::get_surface_density() const
{
	return surface_density;
}

/**
 * returns the radius around a LIGSITEcs pocket grid point that is used to look for
 * residues forming the pocket
 */
const float Parameters::get_pocket_radius() const
{
	return pocket_radius;
}


void Parameters::set_calculate_ligand_clashes(bool b)
{
	calculate_ligand_clashes = b;
}

void Parameters::set_calculate_interaction_centers(bool b)
{

	calculate_interaction_centers = b;
}

