/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef PARAMETERS_H_
#define PARAMETERS_H_

#include <BALL/FORMAT/INIFile.h>
#include <BALL/DATATYPE/string.h>
#include <BALL/KERNEL/iterator.h>
#include <BALL/DATATYPE/stringHashMap.h>
#include <set>

#include "scaffold_selection_utilities.h"
using namespace std;
using namespace BALL;

/**
 * Class providing the global parameters for a ScaffoldSelection run.
 * The parameter values are read from the user-prepared parameter file
 */
class Parameters
{

public:

    /**
     * Constructor
     */
    Parameters();

    /**
     * Default constructor that reads in the parameters from the parameter file
     */
    Parameters(String const& parameter_file, bool verbose, bool time_is_money);

    /**
     * Destructor
     */
    virtual ~Parameters();

    /**
     * returns true if the backbone clash penalty should be computed,
     * false otherwise
     */
    bool get_calculate_backbone_clashes() const;

    /**
     * returns true if the interaction center penalty should be computed,
     * false otherwise
     */
    bool get_calculate_interaction_centers() const;

    /**
     * returns true if the relative SASA penalty should be computed,
     * false otherwise
     */
    bool get_calculate_relative_SAS() const;

    /**
     * returns true if the sphere clash penalty should be computed,
     * false otherwise
     */
    bool get_calculate_ligand_clashes() const;

    /**
     * returns true if the rotamer score should be computed, false otherwise
     */
    bool get_calculate_rotamer_probabilities() const;

    /**
     * returns the name (and path) of the directory containing
     * the PDB structures to be searched
     */
    const String get_structure_directory() const;

    /**
     * returns the minimal number of residues that a protein can have to be
     * eligible as a scaffold
     */
    const unsigned int get_minimal_residue_number() const;

    /**
     * returns the maximal number of residues that a protein can have to be
     * eligible as a scaffold
     */
    const unsigned int get_maximal_residue_number() const;

    /**
     * returns the minimal resolution in Angstroem that a protein must have
     * to be eligible as a scaffold
     */
    const float get_minimal_resolution() const;

    /**
     * returns the minimal number of residues a chain must have in order to be
     * included in the search
     */
    const unsigned int get_minimal_chain_length() const;

    /**
     * returns the number of residues at the boundaries of a beta strand that
     * are eligible as attachment sites
     */
    const unsigned int get_strand_boundary_length() const;

    /**
     * returns the number of residues at the boundaries of an alpha helix
     * that are eligible as attachment sites
     */
    const unsigned int get_helix_boundary_length() const;

    /**
     * returns the maximal number motif hit entries in a ranking list
     */
    const unsigned int get_ranking_list_length() const;

    /**
     * returns the name (and path) of the text file containing the filenames of
     * the PDB files that should be
     * included in the search
     */
    const String get_include_list_file() const;

    /**
     * returns the reference SASA of the argument amino acid in Angstroem^2
     */
    double get_reference_SASA(const String& amino_acid);

    /**
     * returns the location of the BALL installation (necessary to access
     * radius files, the fragment DB etc.)
     */
    const String get_ball_directory() const;

    /**
     * returns true if pockets should be computed and used to determine
     * residues that form them
     */
    bool get_compute_surface_pockets();

    /**
     * returns spacing of the LIGSITEcs grid
     */
    const float get_pocket_grid_space() const;

    /**
     * returns how many of the largest pockets found by LIGSITEcs are
     * used by ScaffoldSelection
     */
    const unsigned int get_number_of_pockets() const;

    /**
     * returns the amount of SSS events a LIGSITEcs grid point has to have
     * in order to be considered as lying in a pocket
     */
    const unsigned int get_SSSthreshold() const;

    /**
     * returns the density of the LIGSITEcs Surface
     */
    const float get_surface_density() const;

    /**
     * returns the radius around a LIGSITEcs pocket grid point
     * that is used to look for residues forming the pocket
     */
    const float get_pocket_radius() const;

    /**
     * sets the calculate_sphere_clashes parameter
     */
    void set_calculate_ligand_clashes(bool b);

    /**
     * sets the calculate_interaction_centers parameter
     */
    void set_calculate_interaction_centers(bool b);

    bool verbose;
    bool time_is_of_the_essence;
private:

    /**
     * true if the backbone clash penalty should be computed, false otherwise
     */
    bool calculate_backbone_clashes;

    /**
     * true if the interaction center penalty should be computed,
     * false otherwise
     */
    bool calculate_interaction_centers;

    /**
     * true if the relative SASA penalty should be computed, false otherwise
     */
    bool calculate_relative_SAS;

    /**
     * true if the sphere clash penalty should be computed, false otherwise
     */
    bool calculate_ligand_clashes;

    /**
     * true if the rotamer score should be computed, false otherwise
     */
    bool calculate_rotamer_probabilities;

    /**
     * name (and path) of the directory containing the PDB structures
     * to be searched
     */
    String structure_directory;

    /**
     * the minimal number of residues that a protein can have to be
     * eligible as a scaffold
     */
    unsigned int minimal_residue_number;

    /**
     * the maximal number of residues that a protein can have to be eligible
     * as a scaffold
     */
    unsigned int maximal_residue_number;

    /**
     * the minimal resolution in Angstroem that a protein must have to be
     * eligible as a scaffold
     */
    float minimal_resolution;

    /**
     * the minimal number of residues a chain must have in order to be
     * included in the search
     */
    unsigned int minimal_chain_length;

    /**
     * The number of residues at the boundaries of an alpha helix that
     * are eligible as attachment sites
     */
    unsigned int helix_boundary_length;

    /**
     * The number of residues at the boundaries of a beta strand that
     * are eligible as attachment sites
     */
    unsigned int strand_boundary_length;

    /**
     * the maximal number motif hit entries in a ranking list
     */
    unsigned int ranking_list_length;

    /**
     * the name (and path) of the text file containing the filenames
     * of the PDB files that should be
     * included in the search
     */
    String include_list_file;

    /**
     * contains the absolute SASA in Angström² of each amino acid type when it
     * is fully exposed to
     * the solvent (measured in a reference peptide GLY-XXX-GLY)
     */
    HashMap<String, double> reference_SAS_areas;

    /**
     * name and path of the file containing the reference SASA values of
     * solvent exposed amino acids
     */
    String reference_SASA_file;

    /**
     * the location of the BALL installation (necessary to access radius
     * files, the fragment DB etc.)
     */
    String ball_directory;

    /**
     * determines whether LIGSITEcs should be used to compute surface pockets
     * in order to exclude
     * residue positions that are not near a pocket from the search for
     * attachment sites
     */
    bool compute_surface_pockets;

    /**
     * the spacing of the grid for the LIGSITEcs pocket finding algorithm
     */
    float pocket_grid_space;

    /**
     * the "number_of_pocket" largest pocket clusters from a LIGSITEcs run
     * will be considered as pockets for
     * the ScaffoldSelection search in the corrensponding structure
     */
    unsigned int number_of_pockets;

    /**
     * determines how many SSS events (Surface-Solvent-Surface) a
     * LIGSITEcs grid point must have to be
     * labeled as being in a pocket
     */
    int SSSthreshold;

    /**
     * the density of vertices in the surface for the LIGSITEcs pocket finding
     * algorithm
     */
    float surface_density;

    /**
     * all residues that have atom(s) that are nearer to a LIGSITEcs
     * pocket grid point than
     * "pocket_radisu" are considered residues that contribute to the
     * geometry of this pocket
     * and are therefore eligible as motif residue attachment positions
     */
    float pocket_radius;

    /**
     * reads in the parameters from the parameter file given as an argument
     */
    void read_parameters(String const &parameter_file);

    /**
     * reads in the reference SAS areas
     */
    void read_reference_SAS_areas();

};

#endif /*PARAMETERS_H_*/
