/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * uses code from LIGSITECS that was modified & adapted for ScaffolSelection
 * by Christoph Malisi under the GPL Gnu Public License
 *
 * The LIGSITECS Copyright notice:
 *      LIGSITECS, a pocket identification method using Connolly surface
 *      http://scoppi.biotec.tu-dresden.de
 *
 *      written by
 *      Bingding Huang
 *      Bioinformatics group
 *      Biotec & Department of Computing
 *      Tazberg 47, 01307
 *      TU Dresden, Germany
 *      Email bhuang@biotec.tu-dresden.de
 *
 *      Copyright (C) LIGSITECS  Bingding Huang
 */


#include "pocketFinder.h"
using namespace BALL;
using namespace std;
using namespace boost;


	PocketFinder::Pocket::Pocket()
	{
		n = 0;
		index=0;
		position = Vector3(0., 0., 0.);
		added=false;
	}

	PocketFinder::Pocket::Pocket(int number, Vector3 pos, int ind)
	{
		n = number;
		position = pos;
		index=ind;
		//neighbour.add(ind);
		added=false;
	}

	PocketFinder::Pocket::~Pocket()
	{
	}

	bool PocketFinder::Pocket::operator < (const Pocket& p) const
	{
		// Note: we implement "<" as ">" on purpose
		// s.t. the largest peak value is on the top of multiset
		return neighbour.size() > p.neighbour.size(); // based on pocket number only
	}


	Vector3 PocketFinder::Pocket::getPosition()
	{
		return position;
	}

	void PocketFinder::Pocket::setPosition(Vector3 pos)
	{
		position = pos;
	}

	void PocketFinder::Pocket::setIndex(int n)
	{
		index=n;

	}

	const int PocketFinder::Pocket::getIndex() const
	{
		return index;
	}



	void PocketFinder::Pocket::addNeighbour(int n)
	{
		if (n!=index)
		{
			neighbour.insert(n);
		}
	}

	HashSet<int> PocketFinder::Pocket::getNeighbour() const
	{
		HashSet<int> tmp=neighbour;
		return tmp;
	}

	const int PocketFinder::Pocket::getSize() const
	{
		return neighbour.size();

	}

	void PocketFinder::Pocket::setAdded()
	{
		added=true;

	}

	bool PocketFinder::Pocket::isAdded()
	{
		return added;
	}

//////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////

PocketFinder::PocketFinder(System &sys, Parameters const &params):
	system(sys)
{
	//system = sys;
	parameters = params;
}

PocketFinder::~PocketFinder()
{
}

Vector3 PocketFinder::getMassCenter()
{
	Vector3 mass_center = Vector3(0., 0., 0.);
	int atom_num = 0;

	for (AtomIterator atom_it = system.beginAtom(); +atom_it; ++atom_it)
	{
		atom_num ++;
		mass_center += atom_it->getPosition();
	}
	mass_center.x = (float)mass_center.x/atom_num;
	mass_center.y = (float)mass_center.y/atom_num;
	mass_center.z = (float)mass_center.z/atom_num;

	return mass_center;
}


Vector3 PocketFinder::getExtent()
{
	Vector3 pos;
	float maxX=0, maxY=0, maxZ=0;
	for (AtomIterator atom_it = system.beginAtom(); +atom_it; ++atom_it)
	{
		pos= atom_it->getPosition();
		if (pos.x<0)
		{
			pos.x =-pos.x;
		}
		if (pos.y<0)
		{
			pos.y=-pos.y;
		}
		if (pos.z<0)
		{
			pos.z=-pos.z;
		}

		if (pos.x>maxX)
		{
			maxX=pos.x;
		}
		if (pos.y>maxY)
		{
			maxY=pos.y;
		}
		if (pos.z>maxZ)
		{
			maxZ=pos.z;
		}
	}
	return Vector3(maxX, maxY, maxZ);
}


Vector3 PocketFinder::getGrid_origin(Vector3 GRID, float GRID_SPACING)
{
	float rx, ry, rz;

	if (int(GRID.x)% 2 == 0)
	{
		rx = GRID.x * GRID_SPACING / 2.0; // Angstrom
	}
	else
	{
		rx = (GRID.x - 1.0) * GRID_SPACING / 2.0; // Angstrom
	}

	if (int(GRID.y)% 2 == 0)
	{
		ry = GRID.y * GRID_SPACING / 2.0; // Angstrom
	}
	else
	{
		ry = (GRID.y - 1.0) * GRID_SPACING / 2.0; // Angstrom
	}

	if (int(GRID.z)% 2 == 0)
	{
		rz = GRID.z * GRID_SPACING / 2.0; // Angstrom
	}
	else
	{
		rz = (GRID.z - 1.0) * GRID_SPACING / 2.0; // Angstrom
	}
	return Vector3(rx, ry, rz); // in unit of Angstroms
}


struct classcomp {
  bool operator() (const vector<unsigned int>& lhs, const vector<unsigned int>& rhs) const
  {return lhs.size() > rhs.size();}
};


HashSet<Residue*> PocketFinder::get_pocket_residues()
{
	float grid_space = parameters.get_pocket_grid_space();
	// set the origin of the protein coordinate system to the mass center of the protein
	Vector3 massCenter = getMassCenter();
	TranslationProcessor tp( -massCenter);
	system.apply(tp);

	// get the extent of the protein coordinates (the largest absolute values of x, y, z of any atom position)
	Vector3 extent = getExtent();
	float x = extent.x*2;
	float y = extent.y*2;
	float z = extent.z*2;

	string radiusFile = parameters.get_ball_directory() + "/data/radii/PARSE.siz"; // the radius file

	AssignRadiusProcessor ar(radiusFile);
	system.apply(ar); //assign the vdW Radii

	int gx = (int)ceil( (x+2.0)/grid_space) + 1;
	int gy = (int)ceil( (y+2.0)/grid_space) + 1;
	int gz = (int)ceil( (z+2.0)/grid_space) + 1;

	Vector3 gc = Vector3(gx, gy, gz);

	Vector3 grid_origin = getGrid_origin(gc, grid_space);

	SurfaceProcessor sp;
	sp.setDensity(parameters.get_surface_density());
	system.apply(sp);
	Surface surface = sp.getSurface();

	multiset<class Pocket> pocket_set;
	vector<GridPoint> grids;

	unsigned int numOfGrid = 0;

	for (int index_x = 0; index_x < gx; index_x++)
	{
		for (int index_y = 0; index_y < gy; index_y++)
		{
			for (int index_z = 0; index_z < gz; index_z++)
			{
				float pos_x = -grid_origin.x +index_x*grid_space;
				float pos_y = -grid_origin.y +index_y*grid_space;
				float pos_z = -grid_origin.z +index_z*grid_space;
				Vector3 pos = Vector3(pos_x, pos_y, pos_z);
				GridPoint g;
				g.position = Vector3(pos_x, pos_y, pos_z);
				g.index = Vector3( (float) index_x, (float) index_y, (float) index_z);
				g.grid_index = numOfGrid;
				g.interactWithProtein = false;
				g.onSurface = false;
				grids.push_back(g);
				numOfGrid++;
			}
		}
	}

	float NEAR_RADIUS = 1.6;
	float WATER_RADIUS = 2.0;


	for (AtomIterator it = system.beginAtom(); +it; ++it)
	{
		Vector3 atom_position = it->getPosition();
		// what is the dimension we have to check for each atom
		// in the units of GRID_SPACING (index)
		int i, j, k;

		int lower_loop_bound_x = (int) ceil((atom_position.x + grid_origin.x
				- it->getRadius() - WATER_RADIUS) /grid_space);
		int upper_loop_bound_x = (int)floor((atom_position.x + grid_origin.x
				+ it->getRadius() + WATER_RADIUS) /grid_space);

		int lower_loop_bound_y = (int) ceil((atom_position.y + grid_origin.y
				- it->getRadius() - WATER_RADIUS) /grid_space);
		int upper_loop_bound_y = (int)floor((atom_position.y + grid_origin.y
				+ it->getRadius() + WATER_RADIUS) /grid_space);

		int lower_loop_bound_z = (int) ceil((atom_position.z + grid_origin.z
				- it->getRadius() - WATER_RADIUS) /grid_space);
		int upper_loop_bound_z = (int)floor((atom_position.z + grid_origin.z
				+ it->getRadius() + WATER_RADIUS) /grid_space);

		// check whether the loop points are out of the global grid bound
		if (lower_loop_bound_x < 0)
		{
			lower_loop_bound_x = 0;
		}
		if (upper_loop_bound_x > gx)
		{
			upper_loop_bound_x = gx;
		}
		if (lower_loop_bound_y < 0)
		{
			lower_loop_bound_y = 0;
		}
		if (upper_loop_bound_y > gy)
		{
			upper_loop_bound_y = gy;
		}
		if (lower_loop_bound_z < 0)
		{
			lower_loop_bound_z = 0;
		}
		if (upper_loop_bound_z > gz)
		{
			upper_loop_bound_z = gz;
		}

		// iterate all the grids around the atom
		// in the units of grid points (index)
		Vector3 pos;

		for (i = lower_loop_bound_x; i < upper_loop_bound_x; i++)
		{
			for (j = lower_loop_bound_y; j < upper_loop_bound_y; j++)
			{
				for (k = lower_loop_bound_z; k < upper_loop_bound_z; k++)
				{
					// set the position we want to check, in the units of Angstrom
					pos=Vector3(-grid_origin.x + i *grid_space, -grid_origin.y +
					            j*grid_space, -grid_origin.z + k*grid_space);
					int index = i*gy*gz + j*gz + k;
					if( !grids[index].interactWithProtein and pos.getDistance(atom_position) <= (it->getRadius()+NEAR_RADIUS) )
					{
						grids[index].interactWithProtein = true;
					}
				}
			}
		}
	}

	Vector3 pos;
	WATER_RADIUS = 1.0;
	float PROBE_RADIUS = 1.0;

	for (unsigned int v=0; v<surface.getNumberOfVertices(); v++)
	{
		Vector3 surface_vertex_position = surface.getVertex(v);

		// what is the dimension we have to check for each surface vertex
		// in the units of GRID_SPACING (index)
		int i, j, k;

		int lower_loop_bound_x=(int) ceil((surface_vertex_position.x
				+ grid_origin.x - WATER_RADIUS) /grid_space);
		int upper_loop_bound_x = (int)floor((surface_vertex_position.x
				+ grid_origin.x + WATER_RADIUS) /grid_space);

		int lower_loop_bound_y = (int) ceil((surface_vertex_position.y
				+ grid_origin.y - WATER_RADIUS) /grid_space);
		int upper_loop_bound_y = (int)floor((surface_vertex_position.y
				+grid_origin.y + WATER_RADIUS) /grid_space);

		int lower_loop_bound_z = (int) ceil((surface_vertex_position.z
				+ grid_origin.z - WATER_RADIUS) /grid_space);
		int upper_loop_bound_z = (int)floor((surface_vertex_position.z
				+ grid_origin.z + WATER_RADIUS) /grid_space);

		// check whether the loop points are out of the global grid bound
		if (lower_loop_bound_x < 0)
		{
			lower_loop_bound_x =0;
		}
		if (upper_loop_bound_x > gx)
		{
			upper_loop_bound_x = gx;
		}
		if (lower_loop_bound_y < 0)
		{
			lower_loop_bound_y = 0;
		}
		if (upper_loop_bound_y > gy)
		{
			upper_loop_bound_y =gy;
		}
		if (lower_loop_bound_z < 0)
		{
			lower_loop_bound_z = 0;
		}
		if (upper_loop_bound_z > gz)
		{
			upper_loop_bound_z = gz;
		}

		// iterate all the grids around the surface vertex
		// in the units of grid points (index)
		Vector3 pos;

		for (i = lower_loop_bound_x; i < upper_loop_bound_x; i++)
		{
			for (j = lower_loop_bound_y; j < upper_loop_bound_y; j++)
			{
				for (k = lower_loop_bound_z; k < upper_loop_bound_z; k++)
				{
					// set the position we want to check, in the units of Angstrom

					pos=Vector3(-grid_origin.x+ i *grid_space, -grid_origin.y+j
					            *grid_space, -grid_origin.z +k *grid_space);
					//pos.set( i *grid_space,+j*grid_space, +k *grid_space );

					int index=i*gy*gz + j*gz + k;

					if ( !grids[index].onSurface and pos.getDistance(surface_vertex_position) <= PROBE_RADIUS )
					{
						grids[index].interactWithProtein = true;
						grids[index].onSurface = true;
					}
				}
			}
		}
	}


	int ix, iy, iz;
	int index_x, index_y, index_z;
	unsigned int index;

	bool touchOneSide = false;
	bool touchBothSide = false;

	unsigned int dialog_len = (unsigned int) ceil(sqrt(float (gz*gz + gy*gy + gx*gx)) ) +1;

	unsigned int pocket_num = 0;

	unsigned int NumberOfSSS;

	vector<GridPoint>::iterator it_g;

	for (it_g=grids.begin(); it_g!=grids.end(); it_g++)
	{
		if (!it_g->interactWithProtein)
		{
			GridPoint g = GridPoint(*it_g);
			ix=(int) g.index.x;
			iy=(int) g.index.y;
			iz=(int) g.index.z;

			NumberOfSSS=0;

			touchOneSide=false;
			touchBothSide=false;

			//printf ("Scan along x--\n ");
			for (index_x=0; index_x<ix; index_x++)
			{
				index=iz + iy*gz + index_x*gz*gy;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				//#//print "Scan along x++ "
				for (index_x=ix+1; index_x<gx; index_x++)
				{
					index=iz + iy*gz + index_x*gz*gy;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;

						break;
					}
				}
			}
			if (touchBothSide)
			{
				NumberOfSSS++;
			}

			//printf ("Scan along y--\n ");
			touchOneSide=false;
			touchBothSide=false;

			for (index_y=0; index_y<iy; index_y++)
			{
				index=iz + index_y*gz + ix*gz*gy;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				//#print "Scan along y++ "
				for (index_y=iy+1; index_y<gy; index_y++)
				{

					index=iz + index_y*gz + ix*gz*gy;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;
						break;
					}
				}
			}
			if (touchBothSide)
			{
				NumberOfSSS++;
			}

			//printf ("Scan along z-- \n");
			touchOneSide=false;
			touchBothSide=false;

			for (index_z=0; index_z<iz; index_z++)
			{
				index=index_z + iy*gz + ix*gz*gy;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				//#print "Scan along z++ "
				for (index_z=iz+1; index_z<gz; index_z++)
				{
					index=index_z + iy*gz + ix*gz*gy;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;
						break;
					}
				}
			}

			if (touchBothSide)
			{
				NumberOfSSS++;

			}

			unsigned int tag;

			int x=ix;
			int y=iy;
			int z=iz;

			touchOneSide=false;
			touchBothSide=false;
			//printf ("Scan dialog x-1, y-1, z+1\n ");

			for (tag=0; tag<dialog_len; tag++)
			{
				x= x- 1;
				y=y-1;
				z=z + 1;
				if ( (x >= gx or x<0 )
						or (y >=gy or y<0 ) or (z >=gz or z<0) )
				{
					break;
				}

				index=x*gy*gz+ y*gz +z;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				x=ix;
				y=iy;
				z=iz;
				// #print "Scan dialog x+1, y+1, z-1, "
				for (tag=0; tag<dialog_len; tag++)
				{
					x= x+ 1;
					y= y + 1;
					z= z - 1;
					if ( (x >= gx or x<0 )
							or (y >=gy or y<0 ) or (z >=gz or z<0) )


					{
						break;
					}

					index=x*gy*gz+ y*gz +z;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;
						break;
					}
				}
			}
			if (touchBothSide)
			{
				NumberOfSSS +=1;
			}

			x=ix;
			y=iy;
			z=iz;

			touchOneSide=false;
			touchBothSide=false;
			//printf ("Scan dialog x-1, y-1, z-1 \n ");
			for (tag=0; tag<dialog_len; tag++)
			{
				x= x- 1;
				y=y-1;
				z=z - 1;
				if ( (x >= gx or x<0 )
						or (y >=gy or y<0 ) or (z >=gz or z<0) )

				{
					break;
				}

				index=x*gy*gz+ y*gz +z;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				x=ix;
				y=iy;
				z=iz;
				// #print "Scan dialog x+1, y+1, z+1, "
				for (tag=0; tag<dialog_len; tag++)
				{
					x= x+ 1;
					y= y + 1;
					z= z + 1;
					if ( (x >= gx or x<0 )
							or (y >=gy or y<0 ) or (z >=gz or z<0) )

					{
						break;
					}

					index=x*gy*gz+ y*gz +z;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;
						break;
					}

				}
			}
			if (touchBothSide)
			{
				NumberOfSSS +=1;
			}

			x=ix;
			y=iy;
			z=iz;

			touchOneSide=false;
			touchBothSide=false;
			//printf ("Scan dialog x+1, y-1, z+1 \n ");
			for (tag=0; tag<dialog_len; tag++)
			{
				x= x + 1;
				y=y - 1;
				z=z + 1;
				if ( (x >= gx or x<0 )
						or (y >=gy or y<0 ) or (z >=gz or z<0) )

				{
					break;
				}

				index=x*gy*gz+ y*gz +z;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				x=ix;
				y=iy;
				z=iz;
				// #print "Scan dialog x-1, y+1, z-1, "
				for (tag=0; tag<dialog_len; tag++)
				{
					x= x - 1;
					y= y + 1;
					z= z - 1;
					if ( (x >= gx or x<0 )
							or (y >=gy or y<0 ) or (z >=gz or z<0) )

					{
						break;
					}

					index=x*gy*gz+ y*gz +z;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;
						break;
					}
				}
			}
			if (touchBothSide)
			{
				NumberOfSSS +=1;
			}

			x=ix;
			y=iy;
			z=iz;

			touchOneSide=false;
			touchBothSide=false;
			//#print "Scan dialog x+1, y-1, z-1, "
			for (tag=0; tag<dialog_len; tag++)
			{
				x= x + 1;
				y=y - 1;
				z=z -1;
				if ( (x >= gx or x<0 )
						or (y >=gy or y<0 ) or (z >=gz or z<0) )

				{
					break;
				}

				index=x*gy*gz+ y*gz +z;
				if (grids[index].onSurface or grids[index].interactWithProtein)
				{
					touchOneSide=true;
					break;
				}
			}

			if (touchOneSide)
			{
				x=ix;
				y=iy;
				z=iz;
				// #print "Scan dialog x-1, y+1, z+1, "
				for (tag=0; tag<dialog_len; tag++)
				{
					x= x - 1;
					y= y + 1;
					z= z + 1;
					if ( (x >= gx or x<0 )
							or (y >=gy or y<0 ) or (z >=gz or z<0) )
					{
						break;
					}

					index = x*gy*gz+ y*gz +z;
					if (grids[index].onSurface or grids[index].interactWithProtein)
					{
						touchBothSide=true;
						break;
					}
				}
			}

			if (touchBothSide)
			{
				NumberOfSSS +=1;
			}
			g.num_vertice = NumberOfSSS;
			//#print "Number of PSP", NumberOfSSS

			// Grid points with enough
			if (NumberOfSSS >= parameters.get_SSSthreshold())
			{
				Pocket p = Pocket(NumberOfSSS, g.position, pocket_num);
				pocket_num +=1;
				pocket_set.insert(p);
			}
			NumberOfSSS=0;
		}
	}


	multiset<class Pocket>::iterator it;

	vector<Pocket> pocket_vec(pocket_set.size());

	unsigned int index_j = 0;

	for (it = pocket_set.begin(); it != pocket_set.end(); it++)
	{
		Pocket p = *it;
		pocket_vec[index_j++] = p;
	}


	int size = pocket_vec.size();

    typedef adjacency_list <vecS, vecS, undirectedS> Graph;

    Graph G;

	// determine which pocket grid points are neighbors
	for (int m = 0; m < size-1; ++m)
	{
		Vector3 pos1 = pocket_vec[m].getPosition();
		for (int n = m+1; n < size; ++n)
		{
			Vector3 pos2 = pocket_vec[n].getPosition();
			float d = pos1.getDistance(pos2);
			if (d <= 3.0 * grid_space)
			{
				add_edge(m,n, G);
			}
		}
	}
	// cluster neighboring pockets
	// by determining the connected components of the neighbour graph
	vector<unsigned int> component(num_vertices(G));
	unsigned int num = connected_components(G, &component[0]);

    set<vector<unsigned int>, classcomp > connected_components;
    for (unsigned int ii = 0; ii != num; ++ii)
    {
    	vector<unsigned int> tmp;
    	for (unsigned int jj = 0; jj < component.size(); ++jj)
    	{
    		if (component[jj] == ii)
    		{
    			tmp.push_back(jj);
    		}
    	}
    	connected_components.insert(tmp);
    }


	HashGrid3<Atom*> hash_grid(-grid_origin, gc, parameters.get_pocket_radius());

	// insert all residues into the grid
	ResidueIterator res_it = system.beginResidue();
	for (; res_it != system.endResidue(); ++res_it)
	{
		AtomIterator at_it = res_it->beginAtom();
		for (; at_it != res_it->endAtom(); ++at_it)
		{
			//insert atoms into the hash grid labeled with the residues
			hash_grid.insert(at_it->getPosition(), &*at_it);
		}
	}
	index_j = 0;

	HashSet<Residue*> pocket_residues;

    set<vector<unsigned int>, classcomp >::iterator setit = connected_components.begin();
    for (; setit != connected_components.end(); ++setit)
	{
    	index_j++;
    	if (index_j > parameters.get_number_of_pockets()) break;

    	vector<unsigned int>::const_iterator it_t = (*setit).begin();
    	for (; it_t != setit->end(); ++it_t)
		{
			Pocket pocket = pocket_vec[*it_t];
			Vector3 pocket_pos = pocket.getPosition();
			HashGridBox3<Atom*>* box = hash_grid.getBox(pocket_pos);
			Position x, y, z;
			hash_grid.getIndices(*box, x, y, z);
			// loop over the neighbour boxes
			for (Index xi = -1; xi <= 1; xi++)
			{
				const Position nx = x + xi;
				for (Index yi = -1; yi <= 1; yi++)
				{
					const Position ny = y + yi;
					for (Index zi = -1; zi <= 1; zi++)
					{
						box = hash_grid.getBox(nx, ny, z+zi);
						if (!box)
						{
							continue;
						}
						if (box->isEmpty())
						{
							continue;
						}
						HashGridBox3<Atom*>::ConstDataIterator bb_it = box->beginData();
						for (; bb_it != box->endData() ; ++bb_it)
						{
							// if the distance between the residue atom and the pocket position is smaller
							// than 3 Angström, remember the Residue
							if (((*bb_it)->getPosition() - pocket_pos).getLength()
									<= parameters.get_pocket_radius())
							{
								pocket_residues.insert((*bb_it)->getResidue());
							}
						}
					}
				}
			}
		}
	}
	// transform back to original coordinates
	tp = TranslationProcessor(massCenter);
	system.apply(tp);

	return pocket_residues;
}
