/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ranking.h"
using namespace BALL;
using namespace std;

Ranking::Ranking(Parameters const &params, String const &mot_name)
{
	parameters = params;
	motif_name = mot_name;
}


Ranking::Ranking()
{
}


Ranking::~Ranking()
{
}



/**
 *
 * @param file
 * @param start
 * @param end
 * @param title
 * @param digit_count
 */
template<typename scorelist_it>
void write_ranking(String const &filename, scorelist_it start,
                   scorelist_it end, unsigned int digit_count = 4)
{
    ofstream file;
    file.open(filename.c_str());

    double digit_mult = pow10(digit_count);
    // loop over the ranking list and write
    // out the scores and the corresponding motif hit identifiers
    for (; start != end; ++start)
    {
        // scores are rounded to 4 digits after the point
        double round_score = start->first * digit_mult;
        round_score = floor(round_score);
        round_score /= digit_mult;
        for(unsigned int i = 0; i < start->second.size(); ++i)
        {
            file << round_score << "\t" << start->second[i] << "\n";
        }
    }
    file << "\n\n";
}


void Ranking::write_rankings ()
{
    write_ranking(motif_name + "_cgp_score.txt",
                  catalytic_atoms_geometry_penalty.begin(),
                  catalytic_atoms_geometry_penalty.end());

    // if backbone clashes were calculated, write the corresponding ranking
    if (parameters.get_calculate_backbone_clashes())
        write_ranking(motif_name + "_backbone_clash_score.txt",
                      backbone_clashes.begin(),
                      backbone_clashes.end());

    // if the sphere clashes were calculated, write the corresponding ranking
    if (parameters.get_calculate_ligand_clashes())
        write_ranking(motif_name + "_substrate_clash_score.txt",
                      ligand_clashes.begin(),
                      ligand_clashes.end());

    // if the interaction center penalties were calculated, write the corresponding ranking
    if (parameters.get_calculate_interaction_centers())
        write_ranking(motif_name + "_ia_center_score.txt",
                      interaction_center_differences.begin(),
                      interaction_center_differences.end());

    // if relative SAS area differences were calculated, write the corresponding ranking
    if (parameters.get_calculate_relative_SAS())
        write_ranking(motif_name + "_relativeSASA_score.txt",
                      relative_SAS_differences.begin(),
                      relative_SAS_differences.end());

    // if the rotamer probabilities were calculated, write the corresponding ranking
    if (parameters.get_calculate_rotamer_probabilities())
        write_ranking(motif_name + "_rotamer_freq_score.txt",
                      rotamer_probabilities.rbegin(),
                      rotamer_probabilities.rend(), 10);

    ofstream file;
    file.open(String(motif_name + "_rotamer_indices.txt").c_str());
    for (StringHashMap<String>::Iterator sit = rotamer_indices.begin();
         sit != rotamer_indices.end(); ++sit)
        file << sit->first << "\t" << sit->second << "\n";

    file.close();

}


/*
 * sorts the motif hits onto ranking lists for each score type, according to the value of the score.
 */
void Ranking::add_motif_hits (list<MotifHit> const &motif_hits)
{

	// loop over all motif hits that were found in this structure
	list<MotifHit>::const_iterator mh_it = motif_hits.begin();
	for (; mh_it != motif_hits.end(); ++mh_it)
	{
		String hit_string = mh_it->get_ID();

		// if the catalytic atoms distances score of the current motif hit is a valid floating number, and
		//  - if either the list is shorter than the maximal length, so the hit will be added no matter what its score,
		//  - or if the score of the hit is better or equal than the worst score on the list,
		// add it to the list
		if ( !isnan(mh_it->get_catalytic_atoms_geometry_penalty()) && !(mh_it->get_catalytic_atoms_geometry_penalty() == INFINITY)
			  && (catalytic_atoms_geometry_penalty.size() < parameters.get_ranking_list_length()
				  || mh_it->get_catalytic_atoms_geometry_penalty() <= catalytic_atoms_geometry_penalty.rbegin()->first) )
		{
			// if there is not yet another hit with the same score on the list, create a vector with the current hit,
			// add it to the list with its score as key
			if (catalytic_atoms_geometry_penalty.find(mh_it->get_catalytic_atoms_geometry_penalty()) == catalytic_atoms_geometry_penalty.end())
			{
				vector<String> temp;
				temp.push_back(hit_string);
				catalytic_atoms_geometry_penalty[mh_it->get_catalytic_atoms_geometry_penalty()] = temp;

				// if the list is now longer than the maximal list length, remove the hit(s)
				// with the worst score from it
				if (catalytic_atoms_geometry_penalty.size() > parameters.get_ranking_list_length())
				{
					catalytic_atoms_geometry_penalty.erase(catalytic_atoms_geometry_penalty.rbegin()->first);
				}
			}
			// if there is already one or more hits with this score on the list, add the current hit tho the corresponding vector
			else
			{
				catalytic_atoms_geometry_penalty[mh_it->get_catalytic_atoms_geometry_penalty()].push_back(hit_string);
			}
		}

		// if the backbone clash score of the current motif hit is a valid floating number, and
		//  - if either the list is shorter than the maximal length, so the hit will be added no matter what its score,
		//  - or if the score of the hit is better or equal than the worst score on the list,
		// add it to the list
		if ( parameters.get_calculate_backbone_clashes() &&
			 !isnan(mh_it->get_backbone_clash_penalty()) && !(mh_it->get_backbone_clash_penalty() == INFINITY)
			 && (backbone_clashes.size() < parameters.get_ranking_list_length()
				 || mh_it->get_backbone_clash_penalty() <= backbone_clashes.rbegin()->first) )
		{
			// if there is not yet another hit with the same score on the list, create a vector with the current hit,
			// add it to the list with its score as key
			if (backbone_clashes.find(mh_it->get_backbone_clash_penalty()) == backbone_clashes.end())
			{
				vector<String> temp;
				temp.push_back(hit_string);
				backbone_clashes[mh_it->get_backbone_clash_penalty()] = temp;

				// if the list is now longer than the maximal list length, remove the hit(s)
				// with the worst score from it
				if (backbone_clashes.size() > parameters.get_ranking_list_length())
				{
					backbone_clashes.erase(backbone_clashes.rbegin()->first);
				}
			}
			// if there is already one or more hits with this score on the list, add the current hit tho the corresponding vector
			else
			{
				backbone_clashes[mh_it->get_backbone_clash_penalty()].push_back(hit_string);
			}
		}

		// if the relative SAS area score of the current motif hit is a valid floating number, and
		//  - if either the list is shorter than the maximal length, so the hit will be added no matter what its score,
		//  - or if the score of the hit is better or equal than the worst score on the list,
		// add it to the list
		if ( parameters.get_calculate_relative_SAS() &&
			 !isnan(mh_it->get_relative_SAS_penalty()) && !(mh_it->get_relative_SAS_penalty() == INFINITY)
			  && (relative_SAS_differences.size() < parameters.get_ranking_list_length()
				  || mh_it->get_relative_SAS_penalty() <= relative_SAS_differences.rbegin()->first) )
		{
			// if there is not yet another hit with the same score on the list, create a vector with the current hit,
			// add it to the list with its score as key
			if (relative_SAS_differences.find(mh_it->get_relative_SAS_penalty()) == relative_SAS_differences.end())
			{
				vector<String> temp;
				temp.push_back(hit_string);
				relative_SAS_differences[mh_it->get_relative_SAS_penalty()] = temp;

				// if the list is now longer than the maximal list length, remove the hit(s)
				// with the worst score from it
				if (relative_SAS_differences.size() > parameters.get_ranking_list_length())
				{
					relative_SAS_differences.erase(relative_SAS_differences.rbegin()->first);
				}
			}
			// if there is already one or more hits with this score on the list, add the current hit tho the corresponding vector
			else
			{
				relative_SAS_differences[mh_it->get_relative_SAS_penalty()].push_back(hit_string);
			}
		}

		// if the rotamer probability score of the current motif hit is a valid floating number, and
		//  - if either the list is shorter than the maximal length, so the hit will be added no matter what its score,
		//  - or if the score of the hit is better or equal than the worst score on the list,
		// add it to the list
		if ( parameters.get_calculate_rotamer_probabilities() &&
			 !isnan(mh_it->get_rotamer_probability()) && !(mh_it->get_rotamer_probability() == INFINITY)
			 && (rotamer_probabilities.size() < parameters.get_ranking_list_length()
				 || mh_it->get_rotamer_probability() >= rotamer_probabilities.begin()->first) )
		{
			// if there is not yet another hit with the same score on the list, create a vector with the current hit,
			// add it to the list with its score as key
			if (rotamer_probabilities.find(mh_it->get_rotamer_probability()) == rotamer_probabilities.end())
			{
				vector<String> temp;
				temp.push_back(hit_string);
				rotamer_probabilities[mh_it->get_rotamer_probability()] = temp;

				// if the list is now longer than the maximal list length, remove the hit(s)
				// with the worst score from it
				if (rotamer_probabilities.size() > parameters.get_ranking_list_length())
				{
					rotamer_probabilities.erase(rotamer_probabilities.begin()->first);
				}
			}
			// if there is already one or more hits with this score on the list, add the current hit tho the corresponding vector
			else
			{
				rotamer_probabilities[mh_it->get_rotamer_probability()].push_back(hit_string);
			}
		}

		// if the sphere clash score of the current motif hit is a valid floating number, and
		//  - if either the list is shorter than the maximal length, so the hit will be added no matter what its score,
		//  - or if the score of the hit is better or equal than the worst score on the list,
		// add it to the list
		if ( parameters.get_calculate_ligand_clashes() &&
			 !isnan(mh_it->get_ligand_clash_penalty()) && !(mh_it->get_ligand_clash_penalty() == INFINITY)
			 && (ligand_clashes.size() < parameters.get_ranking_list_length()
				 || mh_it->get_ligand_clash_penalty() <= ligand_clashes.rbegin()->first) )
		{
			// if there is not yet another hit with the same score on the list, create a vector with the current hit,
			// add it to the list with its score as key
			if (ligand_clashes.find(mh_it->get_ligand_clash_penalty()) == ligand_clashes.end())
			{
				vector<String> temp;
				temp.push_back(hit_string);
				ligand_clashes[mh_it->get_ligand_clash_penalty()] = temp;

				// if the list is now longer than the maximal list length, remove the hit(s)
				// with the worst score from it
				if (ligand_clashes.size() > parameters.get_ranking_list_length())
				{
					ligand_clashes.erase(ligand_clashes.rbegin()->first);
				}
			}
			// if there is already one or more hits with this score on the list, add the current hit tho the corresponding vector
			else
			{
				ligand_clashes[mh_it->get_ligand_clash_penalty()].push_back(hit_string);
			}
		}

		// if the interaction center score of the current motif hit is a valid floating number, and
		//  - if either the list is shorter than the maximal length, so the hit will be added no matter what its score,
		//  - or if the score of the hit is better or equal than the worst score on the list,
		// add it to the list
		if ( parameters.get_calculate_interaction_centers() &&
			 !isnan(mh_it->get_interaction_centers_penalty()) && !(mh_it->get_interaction_centers_penalty() == INFINITY)
			 && (interaction_center_differences.size() < parameters.get_ranking_list_length()
				 || mh_it->get_interaction_centers_penalty() <= interaction_center_differences.rbegin()->first) )
		{
			// if there is not yet another hit with the same score on the list, create a vector with the current hit,
			// add it to the list with its score as key
			if (interaction_center_differences.find(mh_it->get_interaction_centers_penalty()) == interaction_center_differences.end())
			{
				vector<String> temp;
				temp.push_back(hit_string);
				interaction_center_differences[mh_it->get_interaction_centers_penalty()] = temp;

				// if the list is now longer than the maximal list length, remove the hit(s)
				// with the worst score from it
				if (interaction_center_differences.size() > parameters.get_ranking_list_length())
				{
					interaction_center_differences.erase(interaction_center_differences.rbegin()->first);
				}
			}
			// if there is already one or more hits with this score on the list, add the current hit tho the corresponding vector
			else
			{
				interaction_center_differences[mh_it->get_interaction_centers_penalty()].push_back(hit_string);
			}
		}

		rotamer_indices[hit_string] = mh_it->get_rotamer_indices();
	}
}

