/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef RANKING_H_
#define RANKING_H_

#include <BALL/DATATYPE/string.h>
#include <map>
#include "parameters.h"
#include "motifHit.h"

using namespace BALL;
using namespace std;

/**
 * The Ranking lists of the found hits for the enzymatic motif attachment sites
 */
class Ranking
{

public:

	/**
	 * Default constructor, a parameter object is needed
	 */
	Ranking(Parameters const &params, String const &mot_name);

	/**
	 * Constructor
	 */
	Ranking();

	/**
	 * Destructor
	 */
	virtual ~Ranking();

	/**
	 * Writes out the single ranking in an output file to
	 */
	void write_rankings ();

	/**
	 * Sorts the motif hits onto ranking lists for each score type, according to the value of the score.
	 */
	void add_motif_hits (list<MotifHit> const &motif_hits);

private:

    /**
     * The name of the motif this is a ranking for
     */
    String motif_name;

	/**
	 * Object containing the global parameters from the parameter file
	 */
	Parameters parameters;

	/**
	 * Keys: the catayltic atoms geometry penalties,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > catalytic_atoms_geometry_penalty;

	/**
	 * Keys: backbone clash penalties,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > backbone_clashes;

	/**
	 * Keys: sphere clash penalties,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > ligand_clashes;

	/**
	 * Keys: realtive SASA difference penalties,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > relative_SAS_differences;

	/**
	 * Keys: rotamer probability scores,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > rotamer_probabilities;

	/**
	 * Keys: normal distribution of geometric values penalties,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > pair_probabilities;

	/**
	 * Keys: interaction center distance penalties,
	 * Values: unique IDs of the corresponding motif hits
	 */
	map<double, vector<String> > interaction_center_differences;

	/**
	 * Keys: unique IDs of motif hits
	 * Values: String with the indices of the rotamers in the RotamerLibrary
	 *         separated by spaces
	 */
	StringHashMap<String> rotamer_indices;

};

#endif /*RANKING_H_*/
