/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef SCAFFOLDSEARCH_H_
#define SCAFFOLDSEARCH_H_

#include <BALL/FORMAT/PDBFile.h>
#include <dirent.h>

#include "motifSearch.h"
#include "parameters.h"
#include "ranking.h"


/**
 * Contains the methods and variables necessary to perform a search for suitable
 * attachment sites for an enzymatic motif in a set of protein structures
 */
class ScaffoldSearch
{

private:


	/**
	 * Contains the global parameters for the Scaffold Selection run
	 */
	Parameters parameters;

	/**
	 * contains a ordered list for each score, the score values are paired with the
	 * motif hits hat achieve the respective value
	 */
	Ranking ranking;

	/**
	 * all files in the PDB structure directory are stored here
	 */
	vector<String> file_list;

	/**
	 * the motif object that defines what to look for
	 */
	Motif motif;


			/// private methods
	/**
	 * reads in the reference SAS areas, stores them in a map and returns it
	 */
	const map<String, double> get_reference_SAS_areas (String const &filename);

	void subdirectory_recursion(vector<String>& files,
	                            HashSet<String> const &include_list,
	                            String& path);

	/**
	 * returns all filenames in the directory with the name dir
	 */
	vector<String> create_pdb_list ();

	/**
	 * reads in the parameters for the scaffold search
	 */
	void read_parameters (String const &parameter_file);

	/**
	 * time intervall in seconds after which a preliminary ranking should be saved
	 */
	static const unsigned int ranking_time_interval = 200;

public:

	/**
	 * Constructor that reads in the parameters for the scaffold search and stores them in the
	 * corresponding memeber variables. stores the motif in the member variable motif
	 */
	ScaffoldSearch (String const &parameter_file, Motif& motif_ref,
	                bool verbose, bool time_is_money);

	virtual ~ScaffoldSearch ();

	/**
	 * starts a search for suitable scaffolds in all pdb-files in the file list
	 */
	void start_scaffold_search ();


};

#endif /*SCAFFOLDSEARCH_H_*/
