/*
 *  This file is part of ScaffoldSelection, a method for the
 *  identification of potential scaffolds
 *  for the attachment of catalytic residues in a defined geometry.
 *
 *  ScaffoldSelection is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  ScaffoldSelection is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with ScaffoldSelection.  If not, see <http://www.gnu.org/licenses/>.
 */



#ifndef SCAFFOLD_SELECTION_UTILITIES_H_
#define SCAFFOLD_SELECTION_UTILITIES_H_

#include <map>
#include <BALL/KERNEL/residue.h>
#include <BALL/STRUCTURE/structureMapper.h>
#include <BALL/FORMAT/lineBasedFile.h>
#include <BALL/MATHS/analyticalGeometry.h>

using namespace BALL;
using namespace std;

namespace scsel_util
{
    /**
     * constructs the position of the HA1 atom, which is used as a
     * CB subsitute, by assuming an ideal tetrahedral geometry of the atoms
     * N, C, CA and HA1
     */
    Vector3 get_ideal_cb_position_for_gly (Vector3 const &N, Vector3 const &C, Vector3 const &CA);

    /**
     * Constructs two alternative positions for the CA atom of cystein or serine using a tetrahedral geometry.
     * Used to generate two more CA-CB vectors for the computation of the distributions of the geometric values for these two residues
     * otherwise, due to the superimposition on S/OG, CB, CA, only one effective rotamer exists for this computation, and variability is lost
     */
    pair <Vector3, Vector3> get_rotamer_ca_positions_for_ser_cys (Vector3 const &CA, Vector3 const &CB, Vector3 const &O_S_G, bool is_serine);

    /**
     * compute the geometric values of a pair of residues
     * (C-alpha distance, angles Cb1-Ca1-Ca2, Ca1-Ca2-Cb2, torsion angle Cb1-Ca1-Ca2-Cb2)
     */
    //bool compute_geometric_values (Residue* const res1, Residue* const res2, vector<double>& geometric_values);
    vector<double> compute_geometric_values (Vector3 const &CA1, Vector3 const &CB1, Vector3 const &CA2,
                                             Vector3 const &CB2);


    /**
     * returns the corresponding amino acid one letter code for the three letter code argument
     */
    String get_one_letter_code(String const &three_letter_code);


    /**
     * reads in the normal distribution vector from the table file
     * the vector contains 350 probabilities for values from 0.0 to 3.49 .
     * the probability for value x is in vector index x*100
     */
    vector<double> read_normal_distribution_vector(const String &filename);


}

#endif /*SCAFFOLD_SELECTION_UTILITIES_H_*/
